# -*- coding: UTF-8 -*-
#
# This file is part of the zad app
#
# Copyright 2021-2022 Axel Rau
#
# Licensed under GPLv3 or LGPLv3 License, (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.gnu.org/licenses/gpl-3.0
#    https://www.gnu.org/licenses/lgpl-3.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



import ipaddress
import logging
import re

import dns.query, dns.resolver, dns.rdatatype
import dns.exception, dns.rcode, dns.message
import dns.update, dns.tsigkeyring, dns.tsig

from PyQt5 import QtCore

import zad

l = logging.getLogger(__name__)

ddnsUpdate = None

class DdnsUpdate(QtCore.QObject):
    global l
    nsupdate_message = QtCore.pyqtSignal(str)           # lmessages for status bar

    def __init__(self, parent=None):
        super(DdnsUpdate, self).__init__(parent)

        self.master = zad.prefs.master_server
        self.keyring = self._get_ddns_keyring()
        return

    def _get_ddns_keyring(self) -> dns.tsigkeyring:
        """
        Read ddns key and return a key ring for dynamic DNS update
        :return: tsigkeyring
        """
        keyfile = zad.prefs.ddns_key_file
        if not keyfile and self.master:                 # none configured
            return

        key_name = secret = ''
        try:
            with open(keyfile) as kf:
                for line in kf:
                    key_name_match = re.search(r'key\s+"([-a-zA-Z]+)"', line, re.ASCII)
                    if key_name_match:
                        key_name = key_name_match.group(1)
                    secret_match = re.search(r'secret\s+"([=/a-zA-Z0-9]+)"', line, re.ASCII)
                    if secret_match:
                        secret = secret_match.group(1)
        except FileNotFoundError:
            self.error('?ddns key file not found - Check settings')
            return None
        if not (key_name and secret):
            self.error('Can''t parse ddns key file: {}{}'.
                            format('Bad key name ' if not key_name_match else '',
                                   'Bad secret' if not secret_match else ''))
            return None
        else:
            ddns_keyring = dns.tsigkeyring.from_text({key_name: secret})
            return ddns_keyring

    def error(self, message: str):
            l.error(message)
            self.nsupdate_message.emit(message)

    def handle(self, zone: str) -> dns.update.Update:
        """
        Obtain a dynamic DNS update instance
        :param zone: name of to update
        :return: dns.update.Update instance
        """

        if not self.keyring:
            self.error('No valid ddns keyfile configured. No ddns update operation possible')
            return False
        return dns.update.Update(zone,
                             keyring=self.keyring,
                             keyalgorithm=dns.tsig.HMAC_SHA256)


    def delete(self, zone_name, owner_name, rtype, rdata) -> bool:
        """
        Delete one RR
        """
        datatype = dns.rdatatype.from_text(rtype)
        update = self.handle(zone_name)
        if not update:
            return False
        update.delete(owner_name, datatype, rdata)
        response = dns.query.tcp(update, self.master)
        rc = response.rcode()
        if rc != dns.rcode.NOERROR:
            l.error('DNS delete operation failed for zone {} / {} {} {}:\n{} ({})'.
                  format(zone_name, owner_name, rtype, rdata, dns.rcode.to_text(rc), rc))
            return False
        else:
            l.info('[{} {} {} removed from zone {}]'.format(owner_name,  rtype, rdata, zone_name))
            self.nsupdate_message.emit('[{} {} {} removed from zone {}]'.format(owner_name, rtype, rdata, zone_name))
            return True

    def create(self, zone_name, owner_name, ttl, rtype, rdata) -> bool:
        ##datatype = dns.rdatatype.from_text(rtype)
        update = self.handle(zone_name)
        if not update:
            return False
        ##update.add(owner_name, datatype, rdata)
        update.add(owner_name, int(ttl), rtype, rdata)
        response: dns.message.Message = dns.query.tcp(update, self.master)
        rc: dns.rcode = response.rcode()
        if rc != dns.rcode.NOERROR:
            l.error('DNS add operation failed for zone {} / {} {} {}:\n{} ({})'.
                  format(zone_name, ttl, rtype, rdata, dns.rcode.to_text(rc), rc))
            return False
        else:
            l.info('[{} {} {} {} added to zone {}]'.format(owner_name, ttl, rtype, rdata, zone_name))
            self.nsupdate_message.emit('[{} {} {} {} added to zone {}]'.format(owner_name, ttl, rtype, rdata, zone_name))
            return True


def setup():
    global ddnsUpdate
    ddnsUpdate = DdnsUpdate()


import logging

from PyQt5 import QtCore

import zad.common

l = logging.getLogger(__name__)

class Prefs(QtCore.QObject):
    """
    storage manager for settings
    """
    def __init__(self):

        super(Prefs, self).__init__()

        QtCore.QCoreApplication.setOrganizationName("chaos1")
        QtCore.QCoreApplication.setOrganizationDomain("chaos1.de")
        QtCore.QCoreApplication.setApplicationName("zad")
        QtCore.QCoreApplication.setApplicationVersion(zad.get_version())

        self._settings = QtCore.QSettings()

        self._master_server = ''
        self._ddns_key_file = ''
        self._ns_for_axfr = ''
        self._initial_domain = ''
        self._default_ip4_prefix = None
        self._default_ip6_prefix = None
        self._log_file = ''
        self._debug = False
        self._ip4_nets = []
        self._ip6_nets = []
        self._ignored_zones = []

        self._settings.beginGroup('gen')
        keys = self._settings.childKeys()
        self._settings.endGroup()
        if len(keys) < 8:               # do we have something on fallback?
                                        # no - initialize with default values and create fallback
            self._settings.setValue('gen/master_server', '')  # masterServerLineEdit
            self._settings.setValue("gen/ddns_key_file", '')  # ddnsKeyFileLineEdit
            self._settings.setValue("gen/ns_for_axfr", zad.common.IP_XFR_NS)  # serverForZoneTransferLineEdit
            self._settings.setValue("gen/initial_domain", zad.common.INITIAL_DOMAIN)  # initialDomainLineEdit
            self._settings.setValue("gen/default_ip4_prefix", '24')  # defaultPrefixIPv4LineEdit
            self._settings.setValue("gen/default_ip6_prefix", '64')  # defaultPrefixIPv6LineEdit
            self._settings.setValue("gen/log_file", zad.common.DEFAULT_LOG_PATH)  # logfileLineEdit
            self._settings.setValue("gen/debug", 'False')  # debugCheckBox

            self._settings.sync()

        self._read_all_settings()       # sync our vars with fallback

    @property
    def master_server(self):
            return self._master_server
            
    @master_server.setter
    def master_server(self, master_server):
        self._master_server = master_server
        self._settings.setValue('gen/master_server', master_server)
    
    
    @property
    def ddns_key_file(self):
            return self._ddns_key_file
            
    @ddns_key_file.setter
    def ddns_key_file(self, ddns_key_file):
        self._ddns_key_file = ddns_key_file
        self._settings.setValue('gen/ddns_key_file', ddns_key_file)

    
    @property
    def ns_for_axfr(self):
            return self._ns_for_axfr
            
    @ns_for_axfr.setter
    def ns_for_axfr(self, ns_for_axfr):
        self._ns_for_axfr = ns_for_axfr
        self._settings.setValue('gen/ns_for_axfr', ns_for_axfr)

    
    @property
    def initial_domain(self):
            return self._initial_domain
            
    @initial_domain.setter
    def initial_domain(self, initial_domain):
        self._initial_domain = initial_domain
        self._settings.setValue('gen/initial_domain', initial_domain)

    
    @property
    def default_ip4_prefix(self):
            return self._default_ip4_prefix
            
    @default_ip4_prefix.setter
    def default_ip4_prefix(self, default_ip4_prefix):
        self._default_ip4_prefix = default_ip4_prefix
        self._settings.setValue('gen/default_ip4_prefix', str(default_ip4_prefix))

    
    @property
    def default_ip6_prefix(self):
            return self._default_ip6_prefix
            
    @default_ip6_prefix.setter
    def default_ip6_prefix(self, default_ip6_prefix):
        self._default_ip6_prefix = default_ip6_prefix
        self._settings.setValue('gen/default_ip6_prefix', str(default_ip6_prefix))

    
    @property
    def log_file(self):
            return self._log_file
            
    @log_file.setter
    def log_file(self, log_file):
        self._log_file = log_file
        self._settings.setValue('gen/log_file', log_file)

    
    @property
    def debug(self):
            return self._debug
            
    @debug.setter
    def debug(self, debug):
        self._debug = debug
        self._settings.setValue('gen/debug', str(debug))

    
    @property
    def ip4_nets(self):
            return self._ip4_nets
    
    @property
    def ip6_nets(self):
            return self._ip6_nets

    @property
    def ignored_zones(self):
            return self._ignored_zones
            

    def sync(self):
        self._settings.sync()

    def get_net_list(self, name) -> []:
    
        li = []
    
        size: int = self._settings.beginReadArray(name)
        ## print('get:size ' + str(size))
        for i in range(size):
            self._settings.setArrayIndex(i)
            li.append(self._settings.value('net'))
        self._settings.endArray()
    
        setattr(self, '_' + name, li)
        return li
    
    
    def _updateNetList(self, name: str, theList: []):

        ## print('update ' + repr(theList))
        if self._settings.contains(name):
            ## print('update: Removing all')
            self._settings.remove(name)
        theList.sort()
        self._settings.beginWriteArray(name)
        for i in range(len(theList)):
            self._settings.setArrayIndex(i)
            self._settings.setValue('net', theList[i])
            ## print('update: Setting {}'.format(theList[i]))
        self._settings.endArray()
        setattr(self, '_' + name, theList)


    def _read_all_settings(self):
    
        self._master_server = self._settings.value('gen/master_server') 
        self._ddns_key_file = self._settings.value("gen/ddns_key_file")
        self._ns_for_axfr = self._settings.value("gen/ns_for_axfr")
        self._initial_domain = self._settings.value("gen/initial_domain")
        self._default_ip4_prefix = int(self._settings.value("gen/default_ip4_prefix"))
        self._default_ip6_prefix = int(self._settings.value("gen/default_ip6_prefix"))
        self._log_file = self._settings.value("gen/log_file")
        self._debug = self._settings.value("gen/debug").lower() in ('true', 'yes', '1', 't', 'y')

        for name in ('ip4_nets', 'ip6_nets', 'ignored_zones'):
            lst = self.get_net_list(name)
            setattr(self, '_' + name, lst)




import logging, sys

import ipaddress

import dns.rdata, dns.rdataclass, dns.rdataclass, dns.rdatatype
import dns.name, dns.resolver, dns.reversename, dns.exception

from PyQt5 import QtCore, QtWidgets

import zad.common
import zad.models.axfr
import zad.models.main
import zad.models.nsupdate

import zad.views.main

l = None

mw: 'zad.views.main.ZaMainWindow' = None

edit_zone_view = None
zone_view_1 = None
zone_view_2 = None
zone_view_3 = None


def setup(m_window):
    global mw
    mw = m_window


class ZoneView(QtCore.QObject):

    def __init__(self,
                 zoneBox: QtWidgets.QComboBox,
                 netBox: QtWidgets.QComboBox,
                 tabView: QtWidgets.QTableView,
                 parent=None):
        super(ZoneView, self).__init__(parent)

        self.zoneBox: QtWidgets.QComboBox = zoneBox
        self.netBox: QtWidgets.QComboBox = netBox
        self.tabView = tabView
        self.zoneBoxNames = []
        self.netBoxNames = []
        self.zone: zad.models.axfr.Zone = None
        self.net_name = ''              # preserved net name

        self.init_tabView()
        self.connect_signals()

    @QtCore.pyqtSlot(str)
    def zoneBoxSelectionChanged(self, zone_name: str):
        if zone_name:
            self.reload_table(zone_name)

    @QtCore.pyqtSlot(str)
    def netBoxSelectionChanged(self, net_name: str):
        if net_name:
            self.reload_net(net_name)

    @QtCore.pyqtSlot(QtCore.QModelIndex)
    def tableRowDoubleClicked_slot(self, index):
        if index.isValid:
            edit_zone_view.otherDoubleClicked(self, index)

    def addZone(self, zone_name):
        ct = cn = None
        if self.zoneBoxNames:
            ct = self.zoneBox.currentText()
        self.zoneBoxNames.append(zone_name)
        self.zoneBoxNames.sort()
        self.zoneBox.clear()
        self.zoneBox.addItems(self.zoneBoxNames)
        if ct:
            self.zoneBox.setCurrentText(ct)


    def reload_table(self, zone_name):
        model = None
        if zone_name:
            self.zone: zad.models.axfr.Zone = zad.models.axfr.Zone.zoneByName(zone_name)
        if self.zone.type in (zad.common.ZTIP4, zad.common.ZTIP6):                      # a net zone
            self.updateNets()
            if not self.net_name:
                return
            if self.net_name in self.zone.nets:             ## FIXME: addNet changed asynchronously?
                model = zad.models.main.ZoneModel(self.zone.nets[self.net_name].data, True)
        else:
            model = zad.models.main.ZoneModel(self.zone.d)
        self.tabView.setModel(model)
        self.zoneBox.setCurrentText(zone_name)

    def updateNets(self):
        """
        Updates netBoxNames from current zone
        """
        self.netBoxNames = list(self.zone.nets.keys())
        if not self.netBoxNames:
            return
        self.netBoxNames.sort()
        self.netBox.clear()
        self.netBox.addItems(self.netBoxNames)
        self.reload_net(self.netBoxNames[0])

    def reload_net(self, net_name):
        if not self.zone or not net_name or self.zone.type not in (
                                                            zad.common.ZTIP4, zad.common.ZTIP6):
            l.info('%% reload_net: premature return. net_name={}, zone={}, type={}'.format(
                                            net_name,
                                            self.zone.name if self.zone else '',
                                            self.zone.type if self.zone else ''))
            return
        if net_name in self.zone.nets:  ## FIXME: addNet changed asynchronously?
            self.net_name = net_name
            model = zad.models.main.ZoneModel(self.zone.nets[self.net_name].data, True)
            self.tabView.setModel(model)
            self.netBox.setCurrentText(net_name)

    def init_tabView(self):
        self.tabView.setColumnWidth(0, 100)
        self.tabView.setColumnWidth(1, 40)
        self.tabView.setColumnWidth(2, 100)
        hh = self.tabView.horizontalHeader()
        hh.setStretchLastSection(True)
        hh.setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)

    def connect_signals(self):
        self.zoneBox.currentTextChanged.connect(self.zoneBoxSelectionChanged)
        self.netBox.currentTextChanged.connect(self.netBoxSelectionChanged)
        self.tabView.doubleClicked.connect(self.tableRowDoubleClicked_slot)



class ZoneEdit(ZoneView):

    zoneEdit_message = QtCore.pyqtSignal(str)           # lmessages for status bar

    def __init__(self,
                 zoneBox: QtWidgets.QComboBox,
                 netBox: QtWidgets.QComboBox,
                 tabView: QtWidgets.QTableView,
                 parent=None):
        super(ZoneView, self).__init__(parent)

        self.zoneBox: QtWidgets.QComboBox = zoneBox
        self.netBox: QtWidgets.QComboBox = netBox
        self.tabView = tabView
        self.zoneBoxNames = []
        self.netBoxNames = []
        self.zone: zad.models.axfr.Zone = None
        self.zone_name = ''
        self.net_name = ''
        self.tableIndex: QtCore.QModelIndex = None
        
        self.formHost = ''
        self.formName = ''
        self.formttl = ''
        self.forType = ''
        self.formRdata = ''

        self.init_tabView()
        self.clearButtons()

        self.connect_signals()


    @QtCore.pyqtSlot(bool)
    def onMinus(self, checked):
        self.blockSignalsOfForm(True)
        ##self.tableIndex = None
        if zad.models.nsupdate.ddnsUpdate.delete(
            self.zone.name,
            mw.nameAddressEdit.text(),
            mw.typeEdit.text(),
            mw.rdataEdit.toPlainText(),
        ):
            zad.models.axfr.Zone.requestReload(self.zone.name)
        self.blockSignalsOfForm(False)
        self.clearButtons()
        self.clearForm()

    @QtCore.pyqtSlot(bool)
    def onPlus(self, checked):
        if not self.updateHost() or not self.formValid() or not mw.ttlEdit.text():
            zad.app.application.beep()
            return
        self.blockSignalsOfForm(True)
        if zad.models.nsupdate.ddnsUpdate.create(
            self.zone.name,
            mw.nameAddressEdit.text(),
            mw.ttlEdit.text(),
            mw.typeEdit.text(),
            mw.rdataEdit.toPlainText(),
        ):
            zad.models.axfr.Zone.requestReload(self.zone.name)
        self.blockSignalsOfForm(False)
        self.clearButtons()

    @QtCore.pyqtSlot(bool)
    def onReset(self, checked):
        self.reloadForm()

    @QtCore.pyqtSlot(bool)
    def onOK(self, checked):
        if not self.updateHost() or not self.formValid() or not mw.ttlEdit.text():
            zad.app.application.beep()
            return
        self.blockSignalsOfForm(True)
        ##self.tableIndex = None
        self.clearButtons()
        zad.models.nsupdate.ddnsUpdate.delete(
            self.zone.name,
            self.formName,
            self.formType,
            self.formRdata
        )
        zad.models.nsupdate.ddnsUpdate.create(
            self.zone.name,
            mw.nameAddressEdit.text(),
            mw.ttlEdit.text(),
            mw.typeEdit.text(),
            mw.rdataEdit.toPlainText(),
        )
        zad.models.axfr.Zone.requestReload(self.zone.name)

    @QtCore.pyqtSlot(str)
    def onEdited(self, text):
        self.edited()

    @QtCore.pyqtSlot()
    def rdataEdited(self):
        self.edited()

    def edited(self):
        mw.buttonOK.setDisabled(False)
        mw.buttonOK.setDefault(True)
        mw.buttonReset.setDisabled(False)
        mw.buttonP.setDefault(False)
        mw.buttonP.setDisabled(False)
        mw.buttonM.setDisabled(True)

    def formValid(self):
        name = mw.nameAddressEdit.text()
        try:
            n = dns.name.from_text(name)
            if not isinstance(n,  dns.name.Name):
                l.error('Editor: ?Malformed name {}'.format(name))
                self.zoneEdit_message.emit('?Malformed name {}'.format(name))
                return False
            z = str(dns.resolver.zone_for_name(name, tcp=True))
            if z != self.zone.name:
                l.error('?Editor: name not in current zone: {}'.format(name))
                self.zoneEdit_message.emit('?name not in current zone: {}'.format(name))
                return False
            rds = dns.rdataset.Rdataset(
                    dns.rdataclass.IN,
                    dns.rdatatype.from_text(mw.typeEdit.text())
                    )
            if mw.ttlEdit.text():
                rds.update_ttl(int(mw.ttlEdit.text()))
            rd = dns.rdata.from_text(
                            rds.rdclass,
                            rds.rdtype,
                            mw.rdataEdit.toPlainText())
            rds.add(rd)
        except BaseException:
            m = '?Editor: Bad RR: {} {} {}, \n because {} - {}'.format(name,
                                                                           mw.typeEdit.text(),
                                                                           mw.rdataEdit.toPlainText(),
                                                                           sys.exc_info()[0].__name__,
                                                                           str(sys.exc_info()[1]))
            l.error(m)
            self.zoneEdit_message.emit(m)
            return False
        return True

    def updateHost(self):
        if self.zone.type == zad.common.ZTDOM:
            return True
        af = mw.nameAddressEdit.text()
        hf = mw.hostLineEdit.text()
        if af == self.formName and hf == self.formHost:
            return True
        if hf != self.formHost:
            if not hf:                          # changing host to empty field is not allowed
                zad.app.application.beep()
                return False
            addr = self.addrFromHost(hf, self.zone.type, self.net_name)
            if addr:
                reverse_addr = addr.reverse_pointer + '.'
                mw.nameAddressEdit.setText(str(reverse_addr))
                return True
            else:
                return False
        else:
            if not af:                          # changing owner name to empty field is not allowed
                zad.app.application.beep()
                return False
            raddr = self.HostFromReverseAddr(af)
            if raddr:
                mw.hostLineEdit.setText(str(raddr))
                return True
            else:
                return False

    def addrFromHost(self, host, zone_type, net_name):
        """
        Return address as string from host
        """
        try:
            if zone_type == zad.common.ZTIP6:
                h = int(host, 16)
                net: ipaddress.IPv6Network = zad.models.axfr.ip6Nets[net_name]
                if h < 0 or h >= net.num_addresses:
                    raise ValueError('IPv6 Hostnumber out of range')
                else:
                    net_int = int(net.network_address)
                    addr_int = net_int + h
                    addr = ipaddress.IPv6Address(addr_int)
                    return addr
            else:
                h = int(host)
                net: ipaddress.IPv4Network = zad.models.axfr.ip4Nets[net_name]
                if h < 0 or h >= net.num_addresses:
                    raise ValueError('IPv4 Hostnumber out of range')
                else:
                    net_int = int(net.network_address)
                    addr_int = net_int + h
                    addr = ipaddress.IPv4Address(addr_int)
                    return addr
        except (TypeError, ValueError):
            m = '?Editor: Bad host number: "{}", \n because {} - {}'.format(host,
                                                                           sys.exc_info()[0].__name__,
                                                                           str(sys.exc_info()[1]))
            l.error(m)
            self.zoneEdit_message.emit(m)
            zad.app.application.beep()
            return ''
            

    def HostFromReverseAddr(self, reverse_addr):
        try:
            if self.zone.type == zad.common.ZTIP6:
                net = zad.models.axfr.ip6Nets[self.net_name]
                a = dns.reversename.to_address(dns.name.from_text(reverse_addr))
                addr = ipaddress.IPv6Address(a)
                if addr not in net:
                    zad.app.application.beep()
                    return ''
                host = int(addr) & int(net.hostmask)
                return hex(host)[2:]
            else:
                net = zad.models.axfr.ip4Nets[self.net_name]
                a = dns.reversename.to_address(dns.name.from_text(reverse_addr))
                addr = ipaddress.IPv4Address(a)
                if addr not in net:
                    zad.app.application.beep()
                    return ''
                host = int(addr) & int(net.hostmask)
            return host
        except dns.exception.DNSException:
            m = '?Editor: Bad origin IP: "{}", \n because {} - {}'.format(reverse_addr,
                                                                           sys.exc_info()[0].__name__,
                                                                           str(sys.exc_info()[1]))
            l.error(m)
            self.zoneEdit_message.emit(m)
            zad.app.application.beep()
            return ''


    @QtCore.pyqtSlot(str)
    def zoneBoxSelectionChanged(self, zone_name: str):
        if zone_name:
            self.reload_table(zone_name)

    @QtCore.pyqtSlot(str)
    def netBoxSelectionChanged(self, net_name: str):
        if net_name:
            self.reload_net(net_name)

    @QtCore.pyqtSlot(QtCore.QModelIndex)
    def tableRowSelected_slot(self, index):
        self.tableIndex = index
        self.loadForm(index)

    @QtCore.pyqtSlot(QtCore.QModelIndex, QtCore.QModelIndex, )
    def onSelectionChanged(self, index, oldIndex):
        ##print('QItemSelection={}'.format(sel.indexes))
        self.tableIndex = index
        self.loadForm(index)

    def loadForm(self, index):
        self.loadBackup(index)
        self.reloadForm()

    def loadBackup(self, index):
        model = self.tabView.model()
        (self.formName, self.formTtl) = self.owner_name_ttl(model, index)
        if self.zone.type == zad.common.ZTDOM:
            self.formHost = ''
        else:
            i = model.createIndex(index.row(), 0)
            self.formHost = model.data(i, QtCore.Qt.DisplayRole)
        i = model.createIndex(index.row(), 3)
        self.formType = model.data(i, QtCore.Qt.DisplayRole)
        i = model.createIndex(index.row(), 4)
        self.formRdata = model.data(i, QtCore.Qt.DisplayRole)

    def reloadForm(self):
        mw.hostLineEdit.setText(self.formHost)
        mw.nameAddressEdit.setText(self.formName)
        mw.ttlEdit.setText(self.formTtl)
        mw.typeEdit.setText(self.formType)
        mw.rdataEdit.setText(self.formRdata)

        self.clearButtons()
        mw.buttonM.setDisabled(False)
        mw.buttonM.setDefault(True)
        self.blockSignalsOfForm(False)

    def owner_name_ttl(self, model, index) -> (str, str):
        """
        return owner name of RR by index
        """
        row = index.row()
        while row >= 0:
            i = model.createIndex(row, 1)
            name = model.data(i, QtCore.Qt.DisplayRole)
            i = model.createIndex(row, 2)
            ttl = model.data(i, QtCore.Qt.DisplayRole)

            if name:
                return (name, ttl)
            row -= 1
        return ('?','?')

    def clearForm(self):
        mw.hostLineEdit.clear()
        mw.nameAddressEdit.clear()
        mw.ttlEdit.clear()
        mw.typeEdit.clear()
        mw.rdataEdit.clear()

    def blockSignalsOfForm(self, state: bool):
        mw.hostLineEdit.blockSignals(state)
        mw.nameAddressEdit.blockSignals(state)
        mw.ttlEdit.blockSignals(state)
        mw.typeEdit.blockSignals(state)
        mw.rdataEdit.blockSignals(state)

    def clearButtons(self):
        mw.buttonM.setDisabled(True)
        mw.buttonM.setDefault(False)
        mw.buttonP.setDisabled(True)
        mw.buttonP.setDefault(False)
        mw.buttonOK.setDisabled(True)
        mw.buttonOK.setDefault(False)
        mw.buttonReset.setDisabled(True)
        mw.buttonReset.setDefault(False)

    def otherDoubleClicked(self, sender: ZoneView, index):
        other_zone: zad.models.axfr.Zone = zad.models.axfr.Zone.zoneByName(sender.zone.name)
        other_model = sender.tabView.model()
        i = other_model.index(index.row(), 0)
        other_host = other_model.data(i, QtCore.Qt.DisplayRole)

        i = other_model.index(index.row(), 1)
        other_type = other_model.data(i, QtCore.Qt.DisplayRole)

        own_type = mw.typeEdit.text()
        if own_type == 'A':
            if other_zone.type != zad.common.ZTIP4 or other_type != 'PTR':
                zad.app.application.beep()
                return
            other_addr = self.addrFromHost(other_host, other_zone.type, sender.net_name)
            if other_addr:
                mw.rdataEdit.setText(str(other_addr))
        elif own_type == 'AAAA':
            if other_zone.type != zad.common.ZTIP6 or other_type != 'PTR':
                zad.app.application.beep()
                return
            other_addr = self.addrFromHost(other_host, other_zone.type, sender.net_name)
            if other_addr:
                mw.rdataEdit.setText(str(other_addr))


    def selfSelected(self, zone_name, row):
        pass


    def addZone(self, zone_name):
        ct = cn = None
        if self.zoneBoxNames:
            ct = self.zoneBox.currentText()
        if zone_name in self.zoneBoxNames:
            self.zoneBoxNames.remove(zone_name)
            if self.zone.type != zad.common.ZTDOM:
                cn = self.net_name
        self.zoneBoxNames.append(zone_name)
        self.zoneBoxNames.sort()
        self.zoneBox.clear()
        self.zoneBox.addItems(self.zoneBoxNames)
        if ct:
            self.zoneBox.setCurrentText(ct)
        if cn:
            self.netBox.setCurrentText(cn)
        if self.tableIndex and self.tableIndex.isValid:
            self.tabView.selectRow(self.tableIndex.row())
            self.loadForm(self.tableIndex)
    
    def reload_table(self, zone_name):
        model = None
        if zone_name:
            self.zone: zad.models.axfr.Zone = zad.models.axfr.Zone.zoneByName(zone_name)
            if zone_name != self.zone_name:
                self.tableIndex = None
                self.zone_name = zone_name
        if self.zone.type in (zad.common.ZTIP4, zad.common.ZTIP6):                      # a net zone
            self.updateNets()
            if not self.net_name or self.net_name not in self.zone.nets.keys():
                return
            model = zad.models.main.EditZoneModel(self.zone.nets[self.net_name].data, True)
        else:
            model = zad.models.main.EditZoneModel(self.zone.d)
            self.clear_netbox()
        self.tabView.setModel(model)
        self.zoneBox.setCurrentText(zone_name)
        self.clearForm()
        self.clearButtons()

    def reload_net(self, net_name):
        if not self.zone or not net_name or self.zone.type not in (
                                                            zad.common.ZTIP4, zad.common.ZTIP6):
            return
        if net_name in self.zone.nets:  ## FIXME: addNet changed asynchronously?
            if net_name != self.net_name:
                self.tableIndex = None
                self.net_name = net_name
            model = zad.models.main.EditZoneModel(self.zone.nets[self.net_name].data, True)
            self.tabView.setModel(model)
            self.netBox.setCurrentText(net_name)
            self.clearForm()
            self.clearButtons()

    def clear_netbox(self):
        self.netBoxNames = []
        self.net_name = ''
        self.netBox.clear()

    def init_tabView(self):
        self.tabView.setColumnWidth(0, 20)
        self.tabView.setColumnWidth(1, 100)
        self.tabView.setColumnWidth(2, 40)
        self.tabView.setColumnWidth(3, 40)
        self.tabView.setColumnWidth(4, 400)
        hh = self.tabView.horizontalHeader()
        hh.setStretchLastSection(True)
        hh.setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)

    def connect_signals(self):
        self.zoneBox.currentTextChanged.connect(self.zoneBoxSelectionChanged)
        self.netBox.currentTextChanged.connect(self.netBoxSelectionChanged)
        self.tabView.clicked.connect(self.tableRowSelected_slot)
        ##self.tabView.currentChanged.connect(self.onSelectionChanged)

        mw.buttonM.clicked.connect(self.onMinus)
        mw.buttonP.clicked.connect(self.onPlus)
        mw.buttonReset.clicked.connect(self.onReset)
        mw.buttonOK.clicked.connect(self.onOK)

        for field in (mw.hostLineEdit, mw.nameAddressEdit, mw.ttlEdit, mw.typeEdit):
            field.textEdited.connect(self.onEdited)
        mw.rdataEdit.acceptRichText = False
        mw.rdataEdit.textChanged.connect(self.rdataEdited)
        self.blockSignalsOfForm(True)


def zone_loaded(zone_name):
    """
    A new zone has been created"
    """
    global mw, l, edit_zone_view, zone_view_1, zone_view_2, zone_view_3

    if not edit_zone_view:
        edit_zone_view = ZoneEdit(mw.comboBoxMainZone,
                                  mw.comboBoxMainSub,
                                  mw.maintableView)
        edit_zone_view.zoneEdit_message.connect(mw.receive_status_bar_message)  ## FIXME
        l = logging.getLogger(__name__)                                         ## FIXME

    if not zone_view_1:
        zone_view_1 = ZoneView(mw.comboBoxZone_1,
                               mw.comboBoxSub_1,
                               mw.tableView_1)
    if not zone_view_2:
        zone_view_2 = ZoneView(mw.comboBoxZone_2,
                               mw.comboBoxSub_2,
                               mw.tableView_2)
    if not zone_view_3:
        zone_view_3 = ZoneView(mw.comboBoxZone_3,
                               mw.comboBoxSub_3,
                               mw.tableView_3)
    
    z: zad.models.axfr.Zone = zad.models.axfr.Zone.zoneByName(zone_name)
    edit_zone_view.addZone(zone_name)
    if z.type == zad.common.ZTIP6:
        zone_view_3.addZone(zone_name)
    elif z.type == zad.common.ZTIP4:
        zone_view_2.addZone(zone_name)
    elif z.type == zad.common.ZTDOM:
        zone_view_1.addZone(zone_name)



import platform
import asyncio
import logging
from PyQt5 import QtCore, QtWidgets
import zad.pyuic.mainwindow
import zad.models.axfr
import zad.models.main
import zad.models.nsupdate
import zad.models.settings
import zad.views.settings
import zad.views.tables

mainWindow = None
statusBar: QtWidgets.QStatusBar = None

l = logging.getLogger(__name__)

class ZaMainWindow(QtWidgets.QMainWindow,zad.pyuic.mainwindow.Ui_mainWindow):
    set_text_signal = QtCore.pyqtSignal(str)
    def __init__(self, parent=None):
        self.previous_status_text = ''

        super(ZaMainWindow,self).__init__(parent)
        self.setupUi(self)
        self.readSettings()
        zad.views.tables.setup(self)
        self.startThread()

    def startThread(self):
        loop = asyncio.get_event_loop()
        self.asynciothread = zad.models.axfr.RunThread(loop)
        self.thread = QtCore.QThread()
        self.asynciothread.moveToThread(self.thread)

        self.set_text_signal.connect(self.asynciothread.set_text_slot)
        self.asynciothread.axfr_message.connect(self.receive_status_bar_message)
        self.asynciothread.zone_loaded_message.connect(self.receive_zone_loaded_message)
        self.thread.started.connect(self.asynciothread.work)

        self.thread.start()

    @QtCore.pyqtSlot(str)
    def receive_status_bar_message(self, msg):
        statusBar.showMessage('{}{}{}'.format(
            self.previous_status_text,
            ' --- ' if self.previous_status_text else '',
            msg))
        self.previous_status_text = msg

    @QtCore.pyqtSlot(str)
    def receive_zone_loaded_message(self, zone_name):
        return zad.views.tables.zone_loaded(zone_name)
        
    ##def sendtotask(self):
    ##    self.set_text_signal.emit("Task: Configured")

    def closeEvent(self, event: QtCore.QEvent):
        self.writeSettings()
        zad.prefs.sync()
        event.accept()

    def readSettings(self):
        if zad.prefs._settings.contains('mainwindow/size'):
            self.resize(zad.prefs._settings.value('mainwindow/size'))
            self.move(zad.prefs._settings.value('mainwindow/pos'))
            if zad.prefs._settings.contains('mainwindow/box1size'):
                self.box1.resize(zad.prefs._settings.value('mainwindow/box1size'))
                self.box1.move(zad.prefs._settings.value('mainwindow/box1pos'))
                self.box2.resize(zad.prefs._settings.value('mainwindow/box2size'))
                self.box2.move(zad.prefs._settings.value('mainwindow/box2pos'))
                self.box3.resize(zad.prefs._settings.value('mainwindow/box3size'))
                self.box3.move(zad.prefs._settings.value('mainwindow/box3pos'))
                self.mainBox.resize(zad.prefs._settings.value('mainwindow/mainBoxsize'))
                self.mainBox.move(zad.prefs._settings.value('mainwindow/mainBoxpos'))
                self.editBox.resize(zad.prefs._settings.value('mainwindow/editBoxsize'))
                self.editBox.move(zad.prefs._settings.value('mainwindow/editBoxpos'))
        else:
            availableGeometry: QtCore.QRect = self.screen().availableGeometry()
            self.resize(availableGeometry.width() / 3, availableGeometry.height() / 2)
            self.move((availableGeometry.width() - self.width()) / 2,
                        (availableGeometry.height() - self.height()) / 2)

    def writeSettings(self):
        zad.prefs._settings.setValue("mainwindow/size", self.size())
        zad.prefs._settings.setValue("mainwindow/pos", self.pos())
        zad.prefs._settings.setValue("mainwindow/box1size", self.box1.size())
        zad.prefs._settings.setValue("mainwindow/box1pos", self.box1.pos())
        zad.prefs._settings.setValue("mainwindow/box2size", self.box2.size())
        zad.prefs._settings.setValue("mainwindow/box2pos", self.box2.pos())
        zad.prefs._settings.setValue("mainwindow/box3size", self.box3.size())
        zad.prefs._settings.setValue("mainwindow/box3pos", self.box3.pos())
        zad.prefs._settings.setValue("mainwindow/mainBoxsize", self.mainBox.size())
        zad.prefs._settings.setValue("mainwindow/mainBoxpos", self.mainBox.pos())
        zad.prefs._settings.setValue("mainwindow/editBoxsize", self.editBox.size())
        zad.prefs._settings.setValue("mainwindow/editBoxpos", self.editBox.pos())
        

    def exit(self):
        pass

    def about(self):
        QtWidgets.QMessageBox.about(self, None,
                "<p>zad - DNS zone administration tool</p><p>Version {}</p>".format(zad.get_version()))

    def connectActions(self):
        self.actionAbout.triggered.connect(self.about)
        self.actionSettings.triggered.connect(zad.views.settings.showSettings)

    def connectSignals(self):
        zad.models.nsupdate.ddnsUpdate.nsupdate_message.connect(self.receive_status_bar_message)

def setup():
    global mainWindow, statusBar

    mainWindow = ZaMainWindow()
    mainWindow.connectActions()
    mainWindow.setWindowTitle(zad.common.windowHeading)
    if platform.system() == 'FreeBSD':
        mainWindow.setStyleSheet("QTableView, QTableView QWidget, QTextEdit, QLabel { font-size : 9pt } \
                                  QComboBox, QLineEdit { font-size : 10pt }")
    statusBar = mainWindow.statusbar
    mainWindow.readSettings()
    mainWindow.show()

    zad.models.nsupdate.setup()
    mainWindow.connectSignals()

    l.debug('After mw.show')
    return




import platform

import ipaddress
import logging
import pprint
import re
from PyQt5 import QtWidgets, QtCore

import zad.app, zad.common
import zad.pyuic.settings
import zad.models.main
import zad.models.settings



class PrefLineEditView(QtCore.QObject):

    def __init__(self,
                 lineEdit: QtWidgets.QLineEdit,
                 prefName: str,
                 isIP: bool):
        super(PrefLineEditView, self).__init__()
        self.lineEdit: QtWidgets.QLineEdit = lineEdit
        self.prefName: str = prefName
        self.isIP = isIP
        self.preservedText = ''

    def readPref(self):
        self.preservedText = str(getattr(zad.prefs, self.prefName))
        self.lineEdit.setText(self.preservedText)

    def writePref(self):
        self.preservedText = self.lineEdit.text()
        setattr(zad.prefs, self.prefName, self.preservedText)

    def revert(self):
        self.lineEdit.setText(self.preservedText)

    def validateIP(self) -> int:
        if self.isIP and self.lineEdit.text():
            try:
                a = ipaddress.ip_address(self.lineEdit.text())
                return 1
            except:
                return -100
        return 0

class ZaSettingsDialog(QtWidgets.QDialog,zad.pyuic.settings.Ui_settingsTabWidget):
    def __init__(self):
        super(ZaSettingsDialog, self).__init__()
        self.setupUi(self)
        self.readSettings()
        self.tabWidget.setCurrentWidget(self.GeneralTab)
        self.lineEditViews = []
        self.setListViews = []

        self.init_pane_1()
        self.connect_signals()

    @QtCore.pyqtSlot(str)
    def onEdited(self, checked):
        self.OkButton.setDisabled(False)
        self.OkButton.setDefault(True)
        self.revertButton.setDisabled(False)

    @QtCore.pyqtSlot(int)
    def onCheckBoxChanged(self, state):
        self.OkButton.setDisabled(False)
        self.OkButton.setDefault(True)
        self.revertButton.setDisabled(False)

    @QtCore.pyqtSlot(bool)
    def onOK(self, checked):
        ips = 0
        for lineEditView in self.lineEditViews:
            ips += lineEditView.validateIP()
        if ips > 0:
            self.revertButton.setDisabled(True)
            for lineEditView in self.lineEditViews:
                lineEditView.writePref()
            self.preservedDebugState = self.debugCheckBox.isChecked()
            zad.prefs.debug = self.preservedDebugState
            zad.prefs.sync()
        else:
            zad.app.application.beep()

    @QtCore.pyqtSlot(bool)
    def onRevert(self, checked):
        self.OkButton.setDisabled(True)
        self.revertButton.setDisabled(True)
        for lineEditView in self.lineEditViews:
            lineEditView.revert()
        self.debugCheckBox.setChecked(self.preservedDebugState)

    def init_pane_1(self):
        for lineEdit, prefName, isIP in [(self.masterServerLineEdit, 'master_server', True),
                                   (self.ddnsKeyFileLineEdit, 'ddns_key_file', False),
                                   (self.serverForZoneTransferLineEdit, 'ns_for_axfr', True),
                                   (self.initialDomainLineEdit, 'initial_domain', False),
                                   (self.defaultPrefixIPv4LineEdit, 'default_ip4_prefix', False),
                                   (self.defaultPrefixIPv6LineEdit, 'default_ip6_prefix', False),
                                   (self.logfileLineEdit, 'log_file', False)]:
            plf = PrefLineEditView(lineEdit, prefName, isIP)
            self.lineEditViews.append(plf)

        ips = 0
        for prefLineEditView in self.lineEditViews:
            prefLineEditView.readPref()
            ips += prefLineEditView.validateIP()
        if ips <= 0:                # bad ip read from settings file or no ip at all
            self.serverForZoneTransferLineEdit.setText(zad.common.IP_XFR_NS)    # use default
            self.masterServerLineEdit.setText('')                               # and no master server
            self.onOK(True)                                                     # write back to file
        self.OkButton.setDisabled(True)
        self.revertButton.setDisabled(True)
        self.preservedDebugState = zad.prefs.debug
        self.debugCheckBox.setChecked(self.preservedDebugState)

    def connect_signals(self):
        for lineEditView in self.lineEditViews:
            lineEditView.lineEdit.textEdited.connect(self.onEdited)
        self.debugCheckBox.stateChanged.connect(self.onCheckBoxChanged)
        self.revertButton.clicked.connect(self.onRevert)
        self.OkButton.clicked.connect(self.onOK)

    def addSetListView(self, slv):
        self.setListViews.append(slv)

    def readSettings(self):
        if zad.prefs._settings.contains('settings_window/size'):
            self.resize(zad.prefs._settings.value('settings_window/size'))
            self.move(zad.prefs._settings.value('settings_window/pos'))

    def writeSettings(self):
        zad.prefs._settings.setValue("settings_window/size", self.size())
        zad.prefs._settings.setValue("settings_window/pos", self.pos())

    def closeEvent(self, event: QtCore.QEvent):
        self.writeSettings()
        zad.prefs.sync()
        event.accept()

sc: QtCore.QSettings = None
sd: ZaSettingsDialog = None
l = logging.getLogger(__name__)

@QtCore.pyqtSlot()
def showSettings():
    global sd
    sd.show()
    sd.activateWindow()


class SetListsView(QtCore.QObject):

    def __init__(self,
                 listPrefName: str,
                 listWidget: QtWidgets.QListWidget,
                 lineEdit: QtWidgets.QLineEdit,
                 minusButton: QtWidgets.QPushButton,
                 plusButton: QtWidgets.QPushButton,
                 revertButton: QtWidgets.QPushButton,
                 okButton: QtWidgets.QPushButton,
                 
                 parent=None):

        super(SetListsView, self).__init__(parent)

        self.listPrefName = listPrefName
        self.listWidget = listWidget
        self.listWidget.setSortingEnabled(True)
        self.getPrefs()
        if self.prefs:
            self.listWidget.insertItems(0, self.prefs)
            self.listWidget.clearSelection()
        self.lineEdit = lineEdit
        self.minusButton = minusButton
        self.minusButton.setDisabled(True)
        self.plusButton = plusButton
        self.plusButton.setDisabled(True)
        self.revertButton = revertButton
        self.revertButton.setDisabled(True)
        self.okButton = okButton
        self.okButton.setDisabled(True)

        self.connect_signals()
        
        self.preservedText = ''     # while editing an existing list entry
        self.row: int = None
        
    @QtCore.pyqtSlot(QtWidgets.QListWidgetItem)
    def onListItemClicked(self, item):
        if item.text():                         # a row has been selected
            if self.listWidget.selectedItems():
                self.lineEdit.setText(item.text())
                self.preservedText = item.text()
                self.row = self.listWidget.currentRow()
                self.minusButton.setDisabled(False)
                self.minusButton.setDefault(True)
                self.plusButton.setDefault(False)
                self.revertButton.setDisabled(True)
                self.okButton.setDisabled(True)
                self.okButton.setDefault(False)
            else:                                   # no row selected
                self.lineEdit.setText('')
                self.preservedText = ''
                self.minusButton.setDisabled(True)
                self.plusButton.setDisabled(True)
                self.plusButton.setDefault(False)
                self.revertButton.setDisabled(True)
                self.okButton.setDisabled(True)
                self.okButton.setDefault(False)

    @QtCore.pyqtSlot(str)
    def onEdited(self, checked):
        if self.preservedText:      # modifying existing row
            self.plusButton.setDisabled(True)
            self.plusButton.setDefault(False)
            self.minusButton.setDisabled(True)
            self.revertButton.setDisabled(False)

            self.okButton.setDisabled(False)
            self.okButton.setDefault(True)
        else:                       # new row may be added with "+"
            self.minusButton.setDisabled(True)
            self.revertButton.setDisabled(False)
            self.okButton.setDisabled(True)
            self.okButton.setDefault(False)

            self.plusButton.setDisabled(False)
            self.plusButton.setDefault(True)

    @QtCore.pyqtSlot()
    def onEndEdited(self):
        ##self.onOK(False)
        pass

    @QtCore.pyqtSlot(bool)
    def onMinus(self, checked):
        self.delPref(self.preservedText)
        self.listWidget.takeItem(self.listWidget.currentRow())
        self.lineEdit.setText('')
        self.preservedText = ''
        self.minusButton.setDisabled(True)

    @QtCore.pyqtSlot(bool)
    def onPlus(self, checked):
        text = self.lineEdit.text()
        if self.validateNet(text) and (not self.prefs or not text in self.prefs):
            self.addPref(text)
            self.listWidget.addItem(text)
            self.plusButton.setDisabled(True)
        else:
            zad.app.application.beep()

    @QtCore.pyqtSlot(bool)
    def onRevert(self, checked):
        self.lineEdit.setText(self.preservedText)
        self.revertButton.setDisabled(True)
        self.okButton.setDisabled(True)
        self.okButton.setDefault(False)

    @QtCore.pyqtSlot(bool)
    def onOK(self, checked):
        row =self.listWidget.currentRow()
        if row:                                             # may called on defocus w/o current row
            if self.validateNet(self.lineEdit.text()):
                self.setPref(row, self.lineEdit.text())
                self.listWidget.takeItem(self.listWidget.currentRow())
                self.listWidget.addItem(self.lineEdit.text())
                self.lineEdit.setText('')
                self.preservedText = ''
                self.plusButton.setDisabled(True)
                self.minusButton.setDisabled(True)
                self.revertButton.setDisabled(True)
                self.okButton.setDisabled(True)
                self.okButton.setDefault(False)
            else:                                           # validation failed
                zad.app.application.beep()
                
    def connect_signals(self):
        self.listWidget.itemClicked.connect(self.onListItemClicked)
        self.lineEdit.textEdited.connect(self.onEdited)
        self.lineEdit.editingFinished.connect(self.onEndEdited)
        self.minusButton.clicked.connect(self.onMinus)
        self.plusButton.clicked.connect(self.onPlus)
        self.revertButton.clicked.connect(self.onRevert)
        self.okButton.clicked.connect(self.onOK)

    def getPrefs(self):
        self.prefs = zad.prefs.get_net_list(self.listPrefName)

    def addPref(self, value):
        self.getPrefs()
        self.prefs.append(value)
        self.updatePrefs()

    def setPref(self, index, value):
        self.getPrefs()
        self.prefs[index] = value
        self.updatePrefs()

    def delPref(self, value):
        self.getPrefs()
        self.prefs.remove(value)
        self.updatePrefs()

    def updatePrefs(self):
        zad.prefs._updateNetList(self.listPrefName, self.prefs)

    def validateNet(self, netname) -> bool:
        try:
            if self.listPrefName == 'ip4_nets':
                ipaddress.IPv4Network(netname)
                return True
            elif self.listPrefName == 'ip6_nets':
                ipaddress.IPv6Network(netname)
                return True
            else:                                   # a zone name
                if re.search(r'[^a-z0-9.-]', netname):
                    return False
                return True
        except:
            return False


    def printSettings(self, place):
        global sc
        print('Settings of list at {}:\n{}'.format(place, self.getPrefs()))
        print('Settings at {}:\n{}'.format(place, pprint.pprint(sc.as_dict())))


def setup():
    global settingsDialog, sd

    sd = ZaSettingsDialog()

    if platform.system() == 'FreeBSD':
        sd.tabWidget.setStyleSheet("QTabBar::tab:selected { color: white; background-color: blue; }")
    
    sd.addSetListView(SetListsView(
        "ip4_nets",
        sd.iPv4ListWidget,
        sd.iPv4LineEdit,
        sd.iPv4MinusButton,
        sd.iPv4PlusButton,
        sd.iPv4RevertButton,
        sd.iPv4OkButton)
    )

    sd.addSetListView(SetListsView(
        "ip6_nets",
        sd.iPv6ListWidget,
        sd.iPv6LineEdit,
        sd.iPv6MinusButton,
        sd.iPv6PlusButton,
        sd.iPv6RevertButton,
        sd.iPv6OkButton)
    )

    sd.addSetListView(SetListsView(
        "ignored_zones",
        sd.ignoredListWidget,
        sd.ignoredLineEdit,
        sd.ignoredMinusButton,
        sd.ignoredPlusButton,
        sd.ignoredRevertButton,
        sd.ignoredOkButton)
    )




# -*- coding: UTF-8 -*-
#
# This file is part of the zad app
#
# Copyright 2021-2022 Axel Rau
#
# Licensed under GPLv3 or LGPLv3 License, (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.gnu.org/licenses/gpl-3.0
#    https://www.gnu.org/licenses/lgpl-3.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from PyQt5 import QtCore, QtWidgets
import logging
import sys

import zad.models.main
import zad.models.settings
import zad.views.main

application: QtWidgets.QApplication = None
event_loop: QtCore.QEventLoop = None

translator: QtCore.QTranslator = None
locale: QtCore.QLocale = None

def run():

    setup()
    init_translations()

    zad.prefs = zad.models.settings.Prefs()
    zad.views.settings.setup()
    init_logging(zad.prefs.log_file)

    ##zad.models.main.setup()
    zad.views.main.setup()
    if zad.prefs.debug: print('Before application.exec_()')
    sys.exit(application.exec_())


def setup():
    global application

    application = QtWidgets.QApplication(sys.argv)
    event_loop = QtCore.QEventLoop(application)

def init_translations():
    """
    Loads translations for a given input app. If a scaling factor is defined for testing, we use
    our own subclass of QTranslator.
    """
    global application, translator, locale

    ##locale = QLocale(system())                    # set to system locale
    QtCore.QLocale.setDefault(QtCore.QLocale('en')) # ? set to English for now
    locale = QtCore.QLocale()


def init_logging(path:str):
    # info, warning and error goes to file and to console if DEBUG set
    # debug goes to console only (if DEBUG set)
    # Debugging threaded code suppresses console messages: Move to log
    

    logger = logging.getLogger()
    """
    if zad.prefs.debug:
        stdout_formatter = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
        
        stdout_handler = logging.StreamHandler(sys.stdout)
        stdout_handler.setLevel(logging.DEBUG)
        stdout_handler.setFormatter(stdout_formatter)

        logger.addHandler(stdout_handler)

    else:
        logger.setLevel(logging.INFO)
   """
    file_formatter = logging.Formatter('%(asctime)s %(name)s - %(levelname)s - %(message)s', 
                              '%m-%d-%Y %H:%M:%S')

    file_handler = logging.FileHandler(path)  ##FIXME##
    if zad.prefs.debug:
        ##file_handler.setLevel(logging.DEBUG)
        logger.setLevel(logging.DEBUG)
    else:
        ##file_handler.setLevel(logging.INFO)
        logger.setLevel(logging.INFO)
    file_handler.setFormatter(file_formatter)

    logger.addHandler(file_handler)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'zad/Designer/mainwindow.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_mainWindow(object):
    def setupUi(self, mainWindow):
        mainWindow.setObjectName("mainWindow")
        mainWindow.resize(1081, 812)
        mainWindow.setSizeIncrement(QtCore.QSize(10, 10))
        self.centralwidget = QtWidgets.QWidget(mainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.splitter_0 = QtWidgets.QSplitter(self.centralwidget)
        self.splitter_0.setOrientation(QtCore.Qt.Vertical)
        self.splitter_0.setObjectName("splitter_0")
        self.splitter_1 = QtWidgets.QSplitter(self.splitter_0)
        self.splitter_1.setOrientation(QtCore.Qt.Horizontal)
        self.splitter_1.setObjectName("splitter_1")
        self.box1 = QtWidgets.QGroupBox(self.splitter_1)
        self.box1.setObjectName("box1")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.box1)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout()
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.comboBoxZone_1 = QtWidgets.QComboBox(self.box1)
        font = QtGui.QFont()
        font.setFamily("Bitstream Vera Sans")
        font.setPointSize(12)
        self.comboBoxZone_1.setFont(font)
        self.comboBoxZone_1.setObjectName("comboBoxZone_1")
        self.verticalLayout_4.addWidget(self.comboBoxZone_1)
        self.comboBoxSub_1 = QtWidgets.QComboBox(self.box1)
        font = QtGui.QFont()
        font.setFamily("Bitstream Vera Sans")
        font.setPointSize(12)
        self.comboBoxSub_1.setFont(font)
        self.comboBoxSub_1.setObjectName("comboBoxSub_1")
        self.verticalLayout_4.addWidget(self.comboBoxSub_1)
        self.tableView_1 = QtWidgets.QTableView(self.box1)
        font = QtGui.QFont()
        font.setFamily("Bitstream Vera Sans Mono")
        font.setPointSize(11)
        self.tableView_1.setFont(font)
        self.tableView_1.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.tableView_1.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.tableView_1.setObjectName("tableView_1")
        self.verticalLayout_4.addWidget(self.tableView_1)
        self.verticalLayout_5.addLayout(self.verticalLayout_4)
        self.box2 = QtWidgets.QGroupBox(self.splitter_1)
        self.box2.setObjectName("box2")
        self.verticalLayout_7 = QtWidgets.QVBoxLayout(self.box2)
        self.verticalLayout_7.setObjectName("verticalLayout_7")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout()
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.comboBoxZone_2 = QtWidgets.QComboBox(self.box2)
        font = QtGui.QFont()
        font.setFamily("Bitstream Vera Sans")
        font.setPointSize(12)
        self.comboBoxZone_2.setFont(font)
        self.comboBoxZone_2.setObjectName("comboBoxZone_2")
        self.verticalLayout_6.addWidget(self.comboBoxZone_2)
        self.comboBoxSub_2 = QtWidgets.QComboBox(self.box2)
        font = QtGui.QFont()
        font.setFamily("Bitstream Vera Sans")
        font.setPointSize(12)
        self.comboBoxSub_2.setFont(font)
        self.comboBoxSub_2.setObjectName("comboBoxSub_2")
        self.verticalLayout_6.addWidget(self.comboBoxSub_2)
        self.tableView_2 = QtWidgets.QTableView(self.box2)
        font = QtGui.QFont()
        font.setFamily("Bitstream Vera Sans Mono")
        font.setPointSize(11)
        self.tableView_2.setFont(font)
        self.tableView_2.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.tableView_2.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.tableView_2.setObjectName("tableView_2")
        self.verticalLayout_6.addWidget(self.tableView_2)
        self.verticalLayout_7.addLayout(self.verticalLayout_6)
        self.box3 = QtWidgets.QGroupBox(self.splitter_1)
        self.box3.setObjectName("box3")
        self.verticalLayout_9 = QtWidgets.QVBoxLayout(self.box3)
        self.verticalLayout_9.setObjectName("verticalLayout_9")
        self.verticalLayout_8 = QtWidgets.QVBoxLayout()
        self.verticalLayout_8.setObjectName("verticalLayout_8")
        self.comboBoxZone_3 = QtWidgets.QComboBox(self.box3)
        font = QtGui.QFont()
        font.setFamily("Bitstream Vera Sans")
        font.setPointSize(12)
        self.comboBoxZone_3.setFont(font)
        self.comboBoxZone_3.setObjectName("comboBoxZone_3")
        self.verticalLayout_8.addWidget(self.comboBoxZone_3)
        self.comboBoxSub_3 = QtWidgets.QComboBox(self.box3)
        font = QtGui.QFont()
        font.setFamily("Bitstream Vera Sans")
        font.setPointSize(12)
        self.comboBoxSub_3.setFont(font)
        self.comboBoxSub_3.setObjectName("comboBoxSub_3")
        self.verticalLayout_8.addWidget(self.comboBoxSub_3)
        self.tableView_3 = QtWidgets.QTableView(self.box3)
        font = QtGui.QFont()
        font.setFamily("Bitstream Vera Sans Mono")
        font.setPointSize(11)
        self.tableView_3.setFont(font)
        self.tableView_3.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.tableView_3.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.tableView_3.setObjectName("tableView_3")
        self.verticalLayout_8.addWidget(self.tableView_3)
        self.verticalLayout_9.addLayout(self.verticalLayout_8)
        self.splitter_2 = QtWidgets.QSplitter(self.splitter_0)
        self.splitter_2.setOrientation(QtCore.Qt.Horizontal)
        self.splitter_2.setObjectName("splitter_2")
        self.mainBox = QtWidgets.QGroupBox(self.splitter_2)
        self.mainBox.setObjectName("mainBox")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout(self.mainBox)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.verticalLayout_2 = QtWidgets.QVBoxLayout()
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.comboBoxMainZone = QtWidgets.QComboBox(self.mainBox)
        font = QtGui.QFont()
        font.setFamily("Bitstream Vera Sans")
        font.setPointSize(12)
        self.comboBoxMainZone.setFont(font)
        self.comboBoxMainZone.setCurrentText("")
        self.comboBoxMainZone.setObjectName("comboBoxMainZone")
        self.verticalLayout_2.addWidget(self.comboBoxMainZone)
        self.comboBoxMainSub = QtWidgets.QComboBox(self.mainBox)
        font = QtGui.QFont()
        font.setFamily("Bitstream Vera Sans")
        font.setPointSize(12)
        self.comboBoxMainSub.setFont(font)
        self.comboBoxMainSub.setObjectName("comboBoxMainSub")
        self.verticalLayout_2.addWidget(self.comboBoxMainSub)
        self.maintableView = QtWidgets.QTableView(self.mainBox)
        font = QtGui.QFont()
        font.setFamily("Bitstream Vera Sans Mono")
        font.setPointSize(11)
        self.maintableView.setFont(font)
        self.maintableView.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.maintableView.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContentsOnFirstShow)
        self.maintableView.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.maintableView.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)
        self.maintableView.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)
        self.maintableView.setObjectName("maintableView")
        self.maintableView.verticalHeader().setVisible(False)
        self.maintableView.verticalHeader().setHighlightSections(False)
        self.verticalLayout_2.addWidget(self.maintableView)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem)
        self.buttonM = QtWidgets.QPushButton(self.mainBox)
        self.buttonM.setObjectName("buttonM")
        self.horizontalLayout_2.addWidget(self.buttonM)
        self.buttonP = QtWidgets.QPushButton(self.mainBox)
        self.buttonP.setObjectName("buttonP")
        self.horizontalLayout_2.addWidget(self.buttonP)
        self.verticalLayout_2.addLayout(self.horizontalLayout_2)
        self.horizontalLayout_3.addLayout(self.verticalLayout_2)
        self.editBox = QtWidgets.QGroupBox(self.splitter_2)
        self.editBox.setObjectName("editBox")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.editBox)
        self.verticalLayout.setObjectName("verticalLayout")
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setFieldGrowthPolicy(QtWidgets.QFormLayout.ExpandingFieldsGrow)
        self.formLayout.setObjectName("formLayout")
        self.nameAddressLabel = QtWidgets.QLabel(self.editBox)
        self.nameAddressLabel.setObjectName("nameAddressLabel")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.nameAddressLabel)
        self.nameAddressEdit = QtWidgets.QLineEdit(self.editBox)
        font = QtGui.QFont()
        font.setFamily("Bitstream Vera Sans")
        font.setPointSize(12)
        self.nameAddressEdit.setFont(font)
        self.nameAddressEdit.setObjectName("nameAddressEdit")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.nameAddressEdit)
        self.ttlLabel = QtWidgets.QLabel(self.editBox)
        self.ttlLabel.setObjectName("ttlLabel")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.ttlLabel)
        self.ttlEdit = QtWidgets.QLineEdit(self.editBox)
        font = QtGui.QFont()
        font.setFamily("Bitstream Vera Sans")
        font.setPointSize(12)
        self.ttlEdit.setFont(font)
        self.ttlEdit.setObjectName("ttlEdit")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.ttlEdit)
        self.typeLabel = QtWidgets.QLabel(self.editBox)
        self.typeLabel.setObjectName("typeLabel")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.typeLabel)
        self.typeEdit = QtWidgets.QLineEdit(self.editBox)
        font = QtGui.QFont()
        font.setFamily("Bitstream Vera Sans")
        font.setPointSize(12)
        self.typeEdit.setFont(font)
        self.typeEdit.setObjectName("typeEdit")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.typeEdit)
        self.rdataLabel = QtWidgets.QLabel(self.editBox)
        self.rdataLabel.setObjectName("rdataLabel")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.rdataLabel)
        self.rdataEdit = QtWidgets.QTextEdit(self.editBox)
        font = QtGui.QFont()
        font.setFamily("Bitstream Vera Sans Mono")
        font.setPointSize(11)
        self.rdataEdit.setFont(font)
        self.rdataEdit.setObjectName("rdataEdit")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.rdataEdit)
        self.hostLabel = QtWidgets.QLabel(self.editBox)
        self.hostLabel.setObjectName("hostLabel")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.hostLabel)
        self.hostLineEdit = QtWidgets.QLineEdit(self.editBox)
        font = QtGui.QFont()
        font.setFamily("Bitstream Vera Sans")
        font.setPointSize(12)
        self.hostLineEdit.setFont(font)
        self.hostLineEdit.setObjectName("hostLineEdit")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.hostLineEdit)
        self.verticalLayout.addLayout(self.formLayout)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem1)
        self.buttonReset = QtWidgets.QPushButton(self.editBox)
        self.buttonReset.setObjectName("buttonReset")
        self.horizontalLayout.addWidget(self.buttonReset)
        self.buttonOK = QtWidgets.QPushButton(self.editBox)
        self.buttonOK.setObjectName("buttonOK")
        self.horizontalLayout.addWidget(self.buttonOK)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.verticalLayout_3.addWidget(self.splitter_0)
        mainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(mainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1081, 22))
        self.menubar.setObjectName("menubar")
        self.menuZad = QtWidgets.QMenu(self.menubar)
        self.menuZad.setObjectName("menuZad")
        mainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(mainWindow)
        self.statusbar.setObjectName("statusbar")
        mainWindow.setStatusBar(self.statusbar)
        self.actionAbout = QtWidgets.QAction(mainWindow)
        self.actionAbout.setObjectName("actionAbout")
        self.actionSettings = QtWidgets.QAction(mainWindow)
        self.actionSettings.setObjectName("actionSettings")
        self.actionZones = QtWidgets.QAction(mainWindow)
        self.actionZones.setObjectName("actionZones")
        self.actionQuit = QtWidgets.QAction(mainWindow)
        self.actionQuit.setObjectName("actionQuit")
        self.menuZad.addAction(self.actionAbout)
        self.menuZad.addAction(self.actionSettings)
        self.menuZad.addAction(self.actionZones)
        self.menubar.addAction(self.menuZad.menuAction())

        self.retranslateUi(mainWindow)
        QtCore.QMetaObject.connectSlotsByName(mainWindow)

    def retranslateUi(self, mainWindow):
        _translate = QtCore.QCoreApplication.translate
        mainWindow.setWindowTitle(_translate("mainWindow", "MainWindow"))
        self.box1.setTitle(_translate("mainWindow", "Domain-Zone"))
        self.box2.setTitle(_translate("mainWindow", "IPv4-Zone"))
        self.box3.setTitle(_translate("mainWindow", "IPv6-Zone"))
        self.mainBox.setTitle(_translate("mainWindow", "Edit-Zone"))
        self.buttonM.setText(_translate("mainWindow", "-"))
        self.buttonP.setText(_translate("mainWindow", "+"))
        self.editBox.setTitle(_translate("mainWindow", "Editor"))
        self.nameAddressLabel.setText(_translate("mainWindow", "Name/Address"))
        self.ttlLabel.setText(_translate("mainWindow", "TTL"))
        self.typeLabel.setText(_translate("mainWindow", "Type"))
        self.rdataLabel.setText(_translate("mainWindow", "Rdata"))
        self.hostLabel.setText(_translate("mainWindow", "Host"))
        self.buttonReset.setText(_translate("mainWindow", "Reset"))
        self.buttonOK.setText(_translate("mainWindow", "OK"))
        self.menuZad.setTitle(_translate("mainWindow", "zad"))
        self.actionAbout.setText(_translate("mainWindow", "About"))
        self.actionSettings.setText(_translate("mainWindow", "Settings"))
        self.actionZones.setText(_translate("mainWindow", "Zones"))
        self.actionQuit.setText(_translate("mainWindow", "Quit"))



# -*- coding: UTF-8 -*-
#
# This file is part of the zad app
#
# Copyright 2021-2022 Axel Rau
#
# Licensed under GPLv3 or LGPLv3 License, (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.gnu.org/licenses/gpl-3.0
#    https://www.gnu.org/licenses/lgpl-3.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Solution from Flavio Garcia <piraz@candango.org>
from setuptools import setup, find_packages


longdesc = '''
In times of DNSsec, edited zone files interfere with resigning activities of
the nameserver. To avoid inconsistency, zones are maintained by dynamic update
(RFC 2136).

zad provides a GUI for dynamic updates and zone visualisation to make address
and host name editing easy like zone file editing.
'''

setup(
    name="zad",
    version="1.0rc1",
    license="GNU General Public License Version 3",
    description=("A GUI tool for maintaining DNS zones via dynamic update"),
    long_description=longdesc,
    long_description_content_type="text/markdown",
    url="https://codeberg.org/ajr/zad",
    project_urls={
        "Bug Tracker": "https://codeberg.org/ajr/zad/issues",
        "Documentation": "https://zad.readthedocs.io"
    },
    author="Axel Rau>",
    author_email="axel.rau@chaos1.de",
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
        "Environment :: X11 Applications :: Qt",
        "Topic :: Internet :: Name Service (DNS)",
        "Intended Audience :: System Administrators",
    ],
    packages=find_packages(include=['zad', 'zad.*']),
    install_requires=[
        "dnspython==2.1.0",
        "PyQt5==5.15.4",
        "qasync==0.17.0"
    ],
    python_requires='>=3.9',
    entry_points={
        'console_scripts': [
            "zad = zad.app:run",
        ],
    },
)




# -*- coding: UTF-8 -*-
#
# This file is part of the zad app
#
# Copyright 2021-2022 Axel Rau
#
# Licensed under GPLv3 or LGPLv3 License, (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.gnu.org/licenses/gpl-3.0
#    https://www.gnu.org/licenses/lgpl-3.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import zad.models.settings


__author__ = "Axel Rau <axel.rau+zad@chaos1,de>"
__version__ = "1.0rc1"
__licence__ = "GNU General Public License Version 3"


def get_version():
    return __version__


def get_author():
    return __author__.split(" <")[0]


def get_author_email():
    return __author__.split(" <")[1][:-1]


prefs: zad.models.settings.Prefs





# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'zad/Designer/settings.ui'
#
# Created by: PyQt5 UI code generator 5.15.4
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_settingsTabWidget(object):
    def setupUi(self, settingsTabWidget):
        settingsTabWidget.setObjectName("settingsTabWidget")
        settingsTabWidget.resize(389, 373)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(settingsTabWidget)
        self.verticalLayout_2.setContentsMargins(-1, 17, -1, -1)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.tabWidget = QtWidgets.QTabWidget(settingsTabWidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.tabWidget.sizePolicy().hasHeightForWidth())
        self.tabWidget.setSizePolicy(sizePolicy)
        self.tabWidget.setMinimumSize(QtCore.QSize(0, 0))
        self.tabWidget.setBaseSize(QtCore.QSize(365, 349))
        self.tabWidget.setAutoFillBackground(False)
        self.tabWidget.setLocale(QtCore.QLocale(QtCore.QLocale.English, QtCore.QLocale.Germany))
        self.tabWidget.setTabShape(QtWidgets.QTabWidget.Rounded)
        self.tabWidget.setElideMode(QtCore.Qt.ElideNone)
        self.tabWidget.setMovable(False)
        self.tabWidget.setObjectName("tabWidget")
        self.GeneralTab = QtWidgets.QWidget()
        self.GeneralTab.setObjectName("GeneralTab")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.GeneralTab)
        self.verticalLayout.setObjectName("verticalLayout")
        self.formLayout = QtWidgets.QFormLayout()
        self.formLayout.setFieldGrowthPolicy(QtWidgets.QFormLayout.ExpandingFieldsGrow)
        self.formLayout.setObjectName("formLayout")
        self.masterServerLabel = QtWidgets.QLabel(self.GeneralTab)
        self.masterServerLabel.setObjectName("masterServerLabel")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.LabelRole, self.masterServerLabel)
        self.masterServerLineEdit = QtWidgets.QLineEdit(self.GeneralTab)
        self.masterServerLineEdit.setObjectName("masterServerLineEdit")
        self.formLayout.setWidget(0, QtWidgets.QFormLayout.FieldRole, self.masterServerLineEdit)
        self.ddnsKeyFileLabel = QtWidgets.QLabel(self.GeneralTab)
        self.ddnsKeyFileLabel.setObjectName("ddnsKeyFileLabel")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.LabelRole, self.ddnsKeyFileLabel)
        self.ddnsKeyFileLineEdit = QtWidgets.QLineEdit(self.GeneralTab)
        self.ddnsKeyFileLineEdit.setObjectName("ddnsKeyFileLineEdit")
        self.formLayout.setWidget(1, QtWidgets.QFormLayout.FieldRole, self.ddnsKeyFileLineEdit)
        self.serverForZoneTransferLabel = QtWidgets.QLabel(self.GeneralTab)
        self.serverForZoneTransferLabel.setObjectName("serverForZoneTransferLabel")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.LabelRole, self.serverForZoneTransferLabel)
        self.serverForZoneTransferLineEdit = QtWidgets.QLineEdit(self.GeneralTab)
        self.serverForZoneTransferLineEdit.setObjectName("serverForZoneTransferLineEdit")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.FieldRole, self.serverForZoneTransferLineEdit)
        self.initialDomainLabel = QtWidgets.QLabel(self.GeneralTab)
        self.initialDomainLabel.setObjectName("initialDomainLabel")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.LabelRole, self.initialDomainLabel)
        self.initialDomainLineEdit = QtWidgets.QLineEdit(self.GeneralTab)
        self.initialDomainLineEdit.setObjectName("initialDomainLineEdit")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.FieldRole, self.initialDomainLineEdit)
        self.defaultPrefixIPv4Label = QtWidgets.QLabel(self.GeneralTab)
        self.defaultPrefixIPv4Label.setObjectName("defaultPrefixIPv4Label")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.LabelRole, self.defaultPrefixIPv4Label)
        self.defaultPrefixIPv4LineEdit = QtWidgets.QLineEdit(self.GeneralTab)
        self.defaultPrefixIPv4LineEdit.setObjectName("defaultPrefixIPv4LineEdit")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.FieldRole, self.defaultPrefixIPv4LineEdit)
        self.defaultPrefixIPv6Label = QtWidgets.QLabel(self.GeneralTab)
        self.defaultPrefixIPv6Label.setObjectName("defaultPrefixIPv6Label")
        self.formLayout.setWidget(5, QtWidgets.QFormLayout.LabelRole, self.defaultPrefixIPv6Label)
        self.defaultPrefixIPv6LineEdit = QtWidgets.QLineEdit(self.GeneralTab)
        self.defaultPrefixIPv6LineEdit.setObjectName("defaultPrefixIPv6LineEdit")
        self.formLayout.setWidget(5, QtWidgets.QFormLayout.FieldRole, self.defaultPrefixIPv6LineEdit)
        self.logfileLabel = QtWidgets.QLabel(self.GeneralTab)
        self.logfileLabel.setObjectName("logfileLabel")
        self.formLayout.setWidget(6, QtWidgets.QFormLayout.LabelRole, self.logfileLabel)
        self.logfileLineEdit = QtWidgets.QLineEdit(self.GeneralTab)
        self.logfileLineEdit.setObjectName("logfileLineEdit")
        self.formLayout.setWidget(6, QtWidgets.QFormLayout.FieldRole, self.logfileLineEdit)
        self.verticalLayout.addLayout(self.formLayout)
        self.debugCheckBox = QtWidgets.QCheckBox(self.GeneralTab)
        self.debugCheckBox.setObjectName("debugCheckBox")
        self.verticalLayout.addWidget(self.debugCheckBox)
        self.groupBox_4 = QtWidgets.QGroupBox(self.GeneralTab)
        self.groupBox_4.setMinimumSize(QtCore.QSize(0, 0))
        self.groupBox_4.setTitle("")
        self.groupBox_4.setFlat(True)
        self.groupBox_4.setObjectName("groupBox_4")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.groupBox_4)
        self.horizontalLayout.setContentsMargins(-1, 3, -1, 3)
        self.horizontalLayout.setObjectName("horizontalLayout")
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout.addItem(spacerItem)
        self.revertButton = QtWidgets.QPushButton(self.groupBox_4)
        self.revertButton.setObjectName("revertButton")
        self.horizontalLayout.addWidget(self.revertButton)
        self.OkButton = QtWidgets.QPushButton(self.groupBox_4)
        self.OkButton.setDefault(True)
        self.OkButton.setObjectName("OkButton")
        self.horizontalLayout.addWidget(self.OkButton)
        self.verticalLayout.addWidget(self.groupBox_4)
        self.tabWidget.addTab(self.GeneralTab, "")
        self.iPv4Tab = QtWidgets.QWidget()
        self.iPv4Tab.setObjectName("iPv4Tab")
        self.verticalLayout_5 = QtWidgets.QVBoxLayout(self.iPv4Tab)
        self.verticalLayout_5.setObjectName("verticalLayout_5")
        self.verticalLayout_4 = QtWidgets.QVBoxLayout()
        self.verticalLayout_4.setObjectName("verticalLayout_4")
        self.iPv4ListWidget = QtWidgets.QListWidget(self.iPv4Tab)
        self.iPv4ListWidget.setFocusPolicy(QtCore.Qt.ClickFocus)
        self.iPv4ListWidget.setObjectName("iPv4ListWidget")
        self.verticalLayout_4.addWidget(self.iPv4ListWidget)
        self.iPv4LineEdit = QtWidgets.QLineEdit(self.iPv4Tab)
        self.iPv4LineEdit.setObjectName("iPv4LineEdit")
        self.verticalLayout_4.addWidget(self.iPv4LineEdit)
        self.groupBox_7 = QtWidgets.QGroupBox(self.iPv4Tab)
        self.groupBox_7.setMinimumSize(QtCore.QSize(0, 0))
        self.groupBox_7.setTitle("")
        self.groupBox_7.setFlat(True)
        self.groupBox_7.setObjectName("groupBox_7")
        self.horizontalLayout_4 = QtWidgets.QHBoxLayout(self.groupBox_7)
        self.horizontalLayout_4.setContentsMargins(-1, 3, -1, 3)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_4.addItem(spacerItem1)
        self.iPv4MinusButton = QtWidgets.QPushButton(self.groupBox_7)
        self.iPv4MinusButton.setObjectName("iPv4MinusButton")
        self.horizontalLayout_4.addWidget(self.iPv4MinusButton)
        self.iPv4PlusButton = QtWidgets.QPushButton(self.groupBox_7)
        self.iPv4PlusButton.setDefault(True)
        self.iPv4PlusButton.setObjectName("iPv4PlusButton")
        self.horizontalLayout_4.addWidget(self.iPv4PlusButton)
        self.verticalLayout_4.addWidget(self.groupBox_7)
        self.groupBox_6 = QtWidgets.QGroupBox(self.iPv4Tab)
        self.groupBox_6.setMinimumSize(QtCore.QSize(0, 0))
        self.groupBox_6.setTitle("")
        self.groupBox_6.setFlat(True)
        self.groupBox_6.setObjectName("groupBox_6")
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout(self.groupBox_6)
        self.horizontalLayout_3.setContentsMargins(-1, 3, -1, 3)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        spacerItem2 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem2)
        self.iPv4RevertButton = QtWidgets.QPushButton(self.groupBox_6)
        self.iPv4RevertButton.setObjectName("iPv4RevertButton")
        self.horizontalLayout_3.addWidget(self.iPv4RevertButton)
        self.iPv4OkButton = QtWidgets.QPushButton(self.groupBox_6)
        self.iPv4OkButton.setDefault(True)
        self.iPv4OkButton.setObjectName("iPv4OkButton")
        self.horizontalLayout_3.addWidget(self.iPv4OkButton)
        self.verticalLayout_4.addWidget(self.groupBox_6)
        self.verticalLayout_5.addLayout(self.verticalLayout_4)
        self.tabWidget.addTab(self.iPv4Tab, "")
        self.IPv6Tab = QtWidgets.QWidget()
        self.IPv6Tab.setObjectName("IPv6Tab")
        self.verticalLayout_9 = QtWidgets.QVBoxLayout(self.IPv6Tab)
        self.verticalLayout_9.setObjectName("verticalLayout_9")
        self.iPv6ListWidget = QtWidgets.QListWidget(self.IPv6Tab)
        self.iPv6ListWidget.setFocusPolicy(QtCore.Qt.ClickFocus)
        self.iPv6ListWidget.setObjectName("iPv6ListWidget")
        self.verticalLayout_9.addWidget(self.iPv6ListWidget)
        self.iPv6LineEdit = QtWidgets.QLineEdit(self.IPv6Tab)
        self.iPv6LineEdit.setObjectName("iPv6LineEdit")
        self.verticalLayout_9.addWidget(self.iPv6LineEdit)
        self.groupBox_12 = QtWidgets.QGroupBox(self.IPv6Tab)
        self.groupBox_12.setMinimumSize(QtCore.QSize(0, 0))
        self.groupBox_12.setTitle("")
        self.groupBox_12.setFlat(True)
        self.groupBox_12.setObjectName("groupBox_12")
        self.horizontalLayout_9 = QtWidgets.QHBoxLayout(self.groupBox_12)
        self.horizontalLayout_9.setContentsMargins(-1, 3, -1, 3)
        self.horizontalLayout_9.setObjectName("horizontalLayout_9")
        spacerItem3 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_9.addItem(spacerItem3)
        self.iPv6MinusButton = QtWidgets.QPushButton(self.groupBox_12)
        self.iPv6MinusButton.setObjectName("iPv6MinusButton")
        self.horizontalLayout_9.addWidget(self.iPv6MinusButton)
        self.iPv6PlusButton = QtWidgets.QPushButton(self.groupBox_12)
        self.iPv6PlusButton.setDefault(True)
        self.iPv6PlusButton.setObjectName("iPv6PlusButton")
        self.horizontalLayout_9.addWidget(self.iPv6PlusButton)
        self.verticalLayout_9.addWidget(self.groupBox_12)
        self.groupBox_11 = QtWidgets.QGroupBox(self.IPv6Tab)
        self.groupBox_11.setMinimumSize(QtCore.QSize(0, 0))
        self.groupBox_11.setTitle("")
        self.groupBox_11.setFlat(True)
        self.groupBox_11.setObjectName("groupBox_11")
        self.horizontalLayout_8 = QtWidgets.QHBoxLayout(self.groupBox_11)
        self.horizontalLayout_8.setContentsMargins(-1, 3, -1, 3)
        self.horizontalLayout_8.setObjectName("horizontalLayout_8")
        spacerItem4 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_8.addItem(spacerItem4)
        self.iPv6RevertButton = QtWidgets.QPushButton(self.groupBox_11)
        self.iPv6RevertButton.setObjectName("iPv6RevertButton")
        self.horizontalLayout_8.addWidget(self.iPv6RevertButton)
        self.iPv6OkButton = QtWidgets.QPushButton(self.groupBox_11)
        self.iPv6OkButton.setDefault(True)
        self.iPv6OkButton.setObjectName("iPv6OkButton")
        self.horizontalLayout_8.addWidget(self.iPv6OkButton)
        self.verticalLayout_9.addWidget(self.groupBox_11)
        self.tabWidget.addTab(self.IPv6Tab, "")
        self.tab = QtWidgets.QWidget()
        self.tab.setObjectName("tab")
        self.verticalLayout_15 = QtWidgets.QVBoxLayout(self.tab)
        self.verticalLayout_15.setObjectName("verticalLayout_15")
        self.ignoredListWidget = QtWidgets.QListWidget(self.tab)
        self.ignoredListWidget.setFocusPolicy(QtCore.Qt.ClickFocus)
        self.ignoredListWidget.setObjectName("ignoredListWidget")
        self.verticalLayout_15.addWidget(self.ignoredListWidget)
        self.ignoredLineEdit = QtWidgets.QLineEdit(self.tab)
        self.ignoredLineEdit.setObjectName("ignoredLineEdit")
        self.verticalLayout_15.addWidget(self.ignoredLineEdit)
        self.groupBox_20 = QtWidgets.QGroupBox(self.tab)
        self.groupBox_20.setMinimumSize(QtCore.QSize(0, 0))
        self.groupBox_20.setTitle("")
        self.groupBox_20.setFlat(True)
        self.groupBox_20.setObjectName("groupBox_20")
        self.horizontalLayout_17 = QtWidgets.QHBoxLayout(self.groupBox_20)
        self.horizontalLayout_17.setContentsMargins(-1, 3, -1, 3)
        self.horizontalLayout_17.setObjectName("horizontalLayout_17")
        spacerItem5 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_17.addItem(spacerItem5)
        self.ignoredMinusButton = QtWidgets.QPushButton(self.groupBox_20)
        self.ignoredMinusButton.setObjectName("ignoredMinusButton")
        self.horizontalLayout_17.addWidget(self.ignoredMinusButton)
        self.ignoredPlusButton = QtWidgets.QPushButton(self.groupBox_20)
        self.ignoredPlusButton.setDefault(True)
        self.ignoredPlusButton.setObjectName("ignoredPlusButton")
        self.horizontalLayout_17.addWidget(self.ignoredPlusButton)
        self.verticalLayout_15.addWidget(self.groupBox_20)
        self.groupBox_21 = QtWidgets.QGroupBox(self.tab)
        self.groupBox_21.setMinimumSize(QtCore.QSize(0, 0))
        self.groupBox_21.setTitle("")
        self.groupBox_21.setFlat(True)
        self.groupBox_21.setObjectName("groupBox_21")
        self.horizontalLayout_18 = QtWidgets.QHBoxLayout(self.groupBox_21)
        self.horizontalLayout_18.setContentsMargins(-1, 3, -1, 3)
        self.horizontalLayout_18.setObjectName("horizontalLayout_18")
        spacerItem6 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_18.addItem(spacerItem6)
        self.ignoredRevertButton = QtWidgets.QPushButton(self.groupBox_21)
        self.ignoredRevertButton.setObjectName("ignoredRevertButton")
        self.horizontalLayout_18.addWidget(self.ignoredRevertButton)
        self.ignoredOkButton = QtWidgets.QPushButton(self.groupBox_21)
        self.ignoredOkButton.setDefault(True)
        self.ignoredOkButton.setObjectName("ignoredOkButton")
        self.horizontalLayout_18.addWidget(self.ignoredOkButton)
        self.verticalLayout_15.addWidget(self.groupBox_21)
        self.tabWidget.addTab(self.tab, "")
        self.verticalLayout_2.addWidget(self.tabWidget)

        self.retranslateUi(settingsTabWidget)
        self.tabWidget.setCurrentIndex(3)
        QtCore.QMetaObject.connectSlotsByName(settingsTabWidget)

    def retranslateUi(self, settingsTabWidget):
        _translate = QtCore.QCoreApplication.translate
        settingsTabWidget.setWindowTitle(_translate("settingsTabWidget", "Preferences"))
        self.masterServerLabel.setText(_translate("settingsTabWidget", "Master Server"))
        self.ddnsKeyFileLabel.setText(_translate("settingsTabWidget", "ddns Key File"))
        self.serverForZoneTransferLabel.setText(_translate("settingsTabWidget", "Server for Zone Transfer"))
        self.initialDomainLabel.setText(_translate("settingsTabWidget", "Initial Domain"))
        self.defaultPrefixIPv4Label.setText(_translate("settingsTabWidget", "Default Prefix IPv4"))
        self.defaultPrefixIPv6Label.setText(_translate("settingsTabWidget", "Default Prefix IPv6"))
        self.logfileLabel.setText(_translate("settingsTabWidget", "Logfile"))
        self.debugCheckBox.setText(_translate("settingsTabWidget", "Debug Log"))
        self.revertButton.setText(_translate("settingsTabWidget", "Revert"))
        self.OkButton.setText(_translate("settingsTabWidget", "Ok"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.GeneralTab), _translate("settingsTabWidget", "General"))
        self.iPv4MinusButton.setText(_translate("settingsTabWidget", "-"))
        self.iPv4PlusButton.setText(_translate("settingsTabWidget", "+"))
        self.iPv4RevertButton.setText(_translate("settingsTabWidget", "Revert"))
        self.iPv4OkButton.setText(_translate("settingsTabWidget", "Ok"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.iPv4Tab), _translate("settingsTabWidget", "IPv4 Nets"))
        self.iPv6MinusButton.setText(_translate("settingsTabWidget", "-"))
        self.iPv6PlusButton.setText(_translate("settingsTabWidget", "+"))
        self.iPv6RevertButton.setText(_translate("settingsTabWidget", "Revert"))
        self.iPv6OkButton.setText(_translate("settingsTabWidget", "Ok"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.IPv6Tab), _translate("settingsTabWidget", "IPv6 Nets"))
        self.ignoredMinusButton.setText(_translate("settingsTabWidget", "-"))
        self.ignoredPlusButton.setText(_translate("settingsTabWidget", "+"))
        self.ignoredRevertButton.setText(_translate("settingsTabWidget", "Revert"))
        self.ignoredOkButton.setText(_translate("settingsTabWidget", "Ok"))
        self.tabWidget.setTabText(self.tabWidget.indexOf(self.tab), _translate("settingsTabWidget", "Ignored Zones"))



import asyncio
import ipaddress
import logging
import operator
import re
import sys
import time

import dns.query, dns.resolver, dns.zone, dns.versioned, dns.rdataset, dns.rdatatype
import dns.exception
import dns.asyncquery, dns.asyncresolver

from PyQt5 import QtCore

from qasync import QEventLoop

import zad.common

l = logging.getLogger(__name__)

domainZones = {}
ip4Zones = {}
ip6Zones = {}

subdomains = {}
ip4Nets = {}
ip6Nets = {}


class RunThread(QtCore.QThread):
    axfr_message = QtCore.pyqtSignal(str)           # lmessages for status bar
    zone_loaded_message = QtCore.pyqtSignal(str)    # zone fqdn

    def __init__(self, loop: asyncio.AbstractEventLoop, parent=None):
        super(RunThread, self).__init__(parent)
        self.loop = loop
        self.coro_loadZones = loadZones
        # create 1st zone
        z = zad.prefs.initial_domain
        if z[-1] != '.':
            z += '.'
        domainZones[z] = DomainZone(z)

    @QtCore.pyqtSlot(str)
    def set_text_slot(self, txt):
        self.text = txt

    def work(self):
        #run the event loop
        try:
            asyncio.ensure_future(self.coro_loadZones(self), loop=self.loop)
            ##await self.coro_loadZones
            self.loop.run_forever()

        finally:
            print("Disconnect...")

    def send_msg(self, name):
        self.axfr_message.emit(name)

    def send_zone_loaded(self, name):
        self.zone_loaded_message.emit(name)



async def do_axfr(ns, z):
    await dns.asyncquery.inbound_xfr(ns, z)

async def do_zoneName(fqdn):
    z = await dns.asyncresolver.zone_for_name(fqdn, tcp=True)
    return str(z)


# Subclasses, which store per net display data

class Ip4ZadNet(ipaddress.IPv4Network):

    def __init__(self, address, strict=True):
        super(Ip4ZadNet, self).__init__(address, strict)
        self.data = []

class Ip6ZadNet(ipaddress.IPv6Network):

    def __init__(self, address, strict=True):
        super(Ip6ZadNet, self).__init__(address, strict)
        self.data = []


class Zone(object):

    def zoneByName(zone_name):
        if zone_name.endswith('ip6.arpa.'):
            return ip6Zones[zone_name]
        elif zone_name.endswith('in-addr.arpa.'):
            return ip4Zones[zone_name]
        else:
            return domainZones[zone_name]
    
    def requestReload(zone_name):
        z = Zone.zoneByName(zone_name)
        z.valid = False

    def __init__(self, zone_name):
        global domainZones, ip4Zones, ip6Zones

        error = self.init_by_subclass   # bail out if not created as subclass
            
        self.name = zone_name
        if self.name[-1] != '.':
            self.name += '.'

        # add ourselves to globale zone list
        if self.type == zad.common.ZTIP6:
            ip6Zones[self.name] = self
        elif self.type == zad.common.ZTIP4:
            ip4Zones[self.name] = self
        elif self.type == zad.common.ZTDOM:
            domainZones[self.name] = self
        else:
            assert False,'?Wrong zone type "{}" in Zone __init__'.format(self.type)
            
        self.z = None
        self.valid = False
        self.unreachable = False
        self.ignored = self.name in zad.prefs.ignored_zones

        self.getNetsFromPrefs()
        
        self.nets = {}              # netname: ipaddress.IpNetwork object

        if self.name.endswith('ip6.arpa.'):
            self.default_net_mask = int(ipaddress.IPv6Network('1::/{}'.format(
                                                zad.prefs.default_ip6_prefix)).netmask)
            self.default_host_mask = int(ipaddress.IPv6Network('1::/{}'.format(
                                                zad.prefs.default_ip6_prefix)).hostmask)
        else:
            self.default_net_mask = int(Ip4ZadNet('1.0.0.0/{}'.format(
                                                zad.prefs.default_ip4_prefix)).netmask)
            self.default_host_mask = int(Ip4ZadNet('1.0.0.0/{}'.format(
                                                zad.prefs.default_ip4_prefix)).hostmask)


    def data(self, row: int, column: int) -> str:
        if not self.z:
            ## self.loadZone()          # FIXME: missing runner
            return ''
        v = self.d[row][column]
        if not v:
            v = ''
        return str(v)

    async def loadZone(self, runner: RunThread):
        """
        Inherited by Ip4Zone and Ip6Zone
        """

        # get zone data per zone transfer
        
        self.z = dns.zone.Zone(self.name, relativize=False)
        runner.send_msg('Loading {} ...'.format(self.name))
        ns = zad.prefs.ns_for_axfr if zad.prefs.ns_for_axfr else zad.prefs.master_server
        ok = False

        try:
            l.info('[Loading zone {} from NS {}]'.format(self.name, ns))
            await do_axfr(ns, self.z)
            ok = True
        except dns.xfr.TransferError:
            l.error('?loadZone: {} AXFR failed with NS={}'.format(self.name, ns))
            runner.send_msg('?loadZone {} failed. Giving up.'.format(self.name))
        except BaseException:
            l.error('%loadZone: {} Error on AXFR with NS={}, \n because {} - {}'.format(self.name,
                                                                                        ns,
                                                                                        sys.exc_info()[0].__name__,
                                                                                        str(sys.exc_info()[1])))
            runner.send_msg('%loadZone: {} Error on AXFR with NS={}, because {} - {}'.format(self.name,
                                                                                        ns,
                                                                                        sys.exc_info()[0].__name__,
                                                                                        str(sys.exc_info()[1])))

        if not ok:
            l.error('?loadZone {} failed. Giving up.'.format(self.name))
            runner.send_msg('?loadZone {} failed. Giving up.'.format(self.name))
            self.unreachable = True
            return

        err = ''
        try:
            self.z.check_origin()
        except dns.zone.NoSOA:
            err = 'SOA RRset'
        except dns.zone.NoNS:
            err = 'NS RRset'
        except KeyError:
            err = 'origin node'
        if err:
            l.error('% {} has no {}'.format(self.name, err))
            runner.send_msg('% {} has no {}'.format(self.name, err))
            self.unreachable = True

        l.debug('[loadZone: zone={}, AXFR of {} nodes done'.format(self.name, len(self.z.keys())))

        # now parse and store zone data in a nested list (nodes)
        
        prefix = '0x' if self.name.endswith('.ip6.arpa.') else ''  #hex host addresses if IPv6
        nodes = []                                  # [host, rrs]   accumulated here
        first = True
        for k in self.z.keys():
            zn = name = str(k)
            if name == '@':
                name = self.name
            rrs = [['', '', '', '', '', '']]        # name, ttl, type, rdata, host, net
            i = 0                                   # index in rrs
            rrs[i][0] = name
            host = 0
            if not first:
                try:
                    addr = dns.reversename.to_address(dns.name.from_text(name))
                except dns.exception.SyntaxError:
                    l.error('? DNS error: Syntax error while converting reverse name {} to address'.format(name))
                    runner.send_msg('? DNS error: Syntax error while converting reverse name {} to address'.format(name))
                    self.unreachable = True
                    return
                (net, host) = self.addNet(addr)     # net_name, host as int
                l.debug('host={}, addr={}, net={}'.format(host, addr, net))
                if self.type == zad.common.ZTIP4:
                    rrs[i][4] = str(host)           # display IPv4 host
                else:
                    rrs[i][4] = hex(host)[2:]       # display IPv6 host in hex without "0x"
                rrs[i][5] = net
            first = False
            node = self.z[zn]
            for the_rdataset in node:               # rdatasets of current node collected in rrs
                rrs[i][1] = str(the_rdataset.ttl)
                for rdata in the_rdataset:
                    srdatatype = dns.rdatatype.to_text(the_rdataset.rdtype)
                    if srdatatype not in ('RRSIG', 'NSEC', 'NSEC3'):
                        rrs[i][2] = srdatatype
                        srdata = str(rdata)
                        rrs[i][3] = srdata
                        if srdatatype == 'MX':
                            srdata = str(rdata.exchange)
                        elif srdatatype == 'PTR':
                            a = dns.reversename.to_address(k)
                            ##if zad.prefs.debug: print('{}  {}'.format(a, name))
                        elif srdatatype == 'SOA' and srdata.endswith('.arpa.'):
                            a = dns.reversename.to_address(self.name)
                            ##if zad.prefs.debug: print('{}  {}'.format(a, self.name))
                        await self.createZoneFromName(srdatatype, srdata)
                        i = i + 1
                        rrs.append(['', '', '', '', '', ''])
                    ##l.debug('.', end=' ', flush=True)
                ##l.debug('+', end=' ', flush=True)

            if host:                                # host as int for sorting
                nodes.append((host, rrs))           # host, [related rdatasets]
            else:                                   # in apex
                nodes.append((0, rrs))              # origin, [related rdatasets]

        # sort by host number (decimal if IP4 or hex if IP6 [with prefix 0x])
        nodes.sort(key=lambda x: x[0])

        # now sort out net related nodes and store it as row list per net in net.data

        for net_name, net in self.nets.items():
            row = 0  # index in net.data
            net.data = [['', '', '', '', '']]
            first = True
            for node in nodes:
                (sort_host, rrs) = node
                for rr in rrs:
                    if first or rr[5] == net_name:
                        ##l.debug('row={}, rr={}'.format(repr(row), repr(rr)))

                        net.data[row][0] = rr[4]         # host number
                        net.data[row][1] = rr[0]         # origin name
                        net.data[row][2] = rr[1]         # ttl
                        net.data[row][3] = rr[2]         # rtype
                        net.data[row][4] = rr[3]         # rdata

                        net.data.append(['', '', '', '', ''])
                        row += 1
                first = False

            if zad.prefs.debug:
                logZones()
                runner.send_msg('Loading of {}, net {} completed with {} RRs'.format(
                                                                self.name, net_name, len(nodes)))

                l.debug('Loading of {}, net {} completed with {} RRs'.format(
                                                                self.name, net_name, len(nodes)))
        runner.send_zone_loaded(self.name)
        self.valid = True

    # create stub zone from rdata of RRs
    async def createZoneFromName(self, dtype, name):
        global domainZones, ip4Zones, ip6Zones

        if dtype in ('NS', 'MX', 'CNAME', 'DNAME', 'PTR', 'SRV'):   # a domain fqdn
            if name[-1] == '.':  # already absolute?
                fqdn = name
            else:               # no, relative - make absolute
                fqdn = name + '.' + self.name
            ## if len(domainZones.keys()) > zad.common.MAX_ZONES:
                ## return
        elif dtype == 'A':                                # address
            ## if len(ip4Zones.keys()) > zad.common.MAX_ZONES:
                ## return
            if name.count('.') < 3:                 # ignore everything below /9
                l.info('[Not creating reverse zone for {}, because below /9]'.format(name))
                return
            fqdn = dns.reversename.from_address(name)
        elif dtype == 'AAAA':                                # address
            ## if len(ip6Zones.keys()) > zad.common.MAX_ZONES:
                ## return
            if name.count(':') < 3:                 # ignore everything below /17
                l.info('[Not creating reverse zone for {}, because below /17]'.format(name))
                return
            fqdn = dns.reversename.from_address(name)
        else:                                                       # ignore others
            return
        try:
            zoneName = await do_zoneName(fqdn)
            l.debug('createZoneFromName: name={}, fqdn={} OK: zone={}'.format(name, fqdn, zoneName))
        except dns.name.EmptyLabel:
            l.warning('%Empty label: name={}, dtype={}'.format(name, dtype))
            return
        except dns.exception.Timeout:
            l.warning('%{} timed out . . .'.format(fqdn))
            for i in range(zad.common.TIMEOUT_RETRIES):
                time.sleep(zad.common.TIMEOUT_SLEEP)
                try:
                    zoneName = await do_zoneName(fqdn)
                except dns.name.EmptyLabel:
                    l.warning('%Empty label: name={}, dtype={}'.format(name, dtype))
                    return
                except dns.exception.Timeout:
                    l.warning('%{} timed out . . .'.format(fqdn))
                    continue
                except BaseException:
                    l.error('%createZoneFromName/zone_for_name name={}, fqdn={},\n    because {} [{}]'.
                         format(name, fqdn,
                         sys.exc_info()[0].__name__,
                         str(sys.exc_info()[1])))
                    return
            else:
                return
        except BaseException:
            l.error('%createZoneFromName/zone_for_name name={}, fqdn={},\n    because {} [{}]'.
                 format(name, fqdn,
                 sys.exc_info()[0].__name__,
                 str(sys.exc_info()[1])))
            return
        ##l.debug('type={}, name={}, fqdn={}, zone={}'.format(dtype, name, fqdn, zoneName))
        l.debug('createZoneFromName: {}'.format(fqdn))
        if dtype == 'A' and zoneName not in ip4Zones:
            ip4Zones[zoneName] = Ip4Zone(zoneName)
        elif dtype == 'AAAA' and zoneName not in ip6Zones:
            ip6Zones[zoneName] = Ip6Zone(zoneName)
        elif ( dtype in ('NS', 'MX', 'CNAME', 'DNAME', 'PTR', 'SRV')  and zoneName not in domainZones ):
            if zoneName.endswith('.arpa.'):
                l.error('?Unexpected reverse address (instead of name) {} in {} RR - ignored'.format(name, dtype))
                return

            domainZones[zoneName] = DomainZone(zoneName)
        if zad.prefs.debug: print('createZoneFromName: {} done'.format(fqdn))

    def getNetsFromPrefs(self):
        if not ip4Nets and zad.prefs.ip4_nets:
            for net_name in zad.prefs.ip4_nets:
                net = ipaddress.IPv4Network(net_name)
                ip4Nets[str(net)] = net            # use normalized form 2a05:bec0:26:: instead of 2a05:bec0:0026::
        if not ip6Nets and zad.prefs.ip6_nets:
            for net_name in zad.prefs.ip6_nets:
                net = ipaddress.IPv6Network(net_name)
                ip6Nets[str(net)] = net

    def addNet(self, address) -> (str, str):
        """
        return tuple of net as string and host as int
        """

        if ':' in address:                              # IPv6
            if self.type == zad.common.ZTIP4:
                l.error('?IPv6 address {} in IPv4 zone {} - ignored'.format(address, self.name))
                l.runner.send_msg('?IPv6 address {} in IPv4 zone {} - ignored'.format(address, self.name))
                return ('', 0)
            
            # try known nets of current zone
            a = ipaddress.IPv6Address(address)
            for k, v in self.nets.items():              # { netname: netobject }
                if a in v:
                    return (k,
                            int(ipaddress.IPv6Address(int(a) & int(v.hostmask))))
            # try all known nets, included those from settings
            for k, v in ip6Nets.items():                # { netname: netobject } but netname may be none-canonical
                if a in v:
                    self.nets[str(v)] = v               # none-canonical: 2a02:cd04:01:: instead of 2a02:cd04:1::
                    return (str(v),
                            int(ipaddress.IPv6Address(int(a) & int(v.hostmask))))
            # now create a new net from default prefix
            try:
                n = Ip6ZadNet((int(a) & self.default_net_mask, int(zad.prefs.default_ip6_prefix)))
            except ValueError:
                l.error('?ValueError: IPv6 address {} in IPv6 zone {}, with mask {}'.format(
                                                                address, self.name, hex(self.default_net_mask)))
                l.runner.send_msg('?ValueError: IPv6 address {} in IPv6 zone {}, with mask {}'.format(
                                                                address, self.name, hex(self.default_net_mask)))
                return ('', 0)
            self.nets[str(n)] = n
            ip6Nets[str(n)] = n
            return (str(n),
                    int(ipaddress.IPv6Address(int(a) & int(n.hostmask))))

        elif '.' in address:
            def trimmHost(addr, net):
                m = re.match(r'^(0\.)+(.*)', str(ipaddress.IPv4Address(int(addr) & int(net.hostmask))), re.A)
                return int(m.group(2))

            if self.type == zad.common.ZTIP6:
                l.error('?IPv4 address {} in IPv6 zone {} - ignored'.format(address, self.name))
                l.runner.send_msg('?IPv4 address {} in IPv6 zone {} - ignored'.format(address, self.name))
                return ('', 0)
                
            a = ipaddress.IPv4Address(address)
            for k, v in self.nets.items():              # { netname: netobject }
                if a in v:
                    return (k,
                            trimmHost(a, v))
            for k, v in ip4Nets.items():                # { netname: netobject }
                if a in v:
                    self.nets[k] = v
                    return (k,
                            trimmHost(a, v))
            n = Ip4ZadNet((int(a) & self.default_net_mask, int(zad.prefs.default_ip4_prefix)))
            self.nets[str(n)] = n
            ip4Nets[str(n)] = n
            return (str(n),
                    trimmHost(a, n))

        else:
            assert False, 'Zone.addNet received invalid address "{}"'.format(address)
            self.unreachable = True


class DomainZone(Zone):
    def __init__(self, zone_name):
        self.init_by_subclass = True
        self.type = zad.common.ZTDOM
        super(DomainZone, self).__init__(zone_name)
        self.d = [['', '', '', '', '']]

    def data(self, row: int, column: int) -> str:
        if not self.z:
            self.loadZone()
        v = self.d[row][column]
        if not v: v = ''
        return str(v)

    async def loadZone(self, runner: RunThread):
        row = 0
        self.d = [['', '', '', '', '']]
        self.z = dns.zone.Zone(self.name, relativize=False)
        runner.send_msg('Loading {} ...'.format(self.name))
        ns = zad.prefs.ns_for_axfr if zad.prefs.ns_for_axfr else zad.prefs.master_server
        ok = False

        try:
            l.info('[Loading zone {} from NS {}]'.format(self.name, ns))
            await do_axfr(ns, self.z)
            ok = True
        except dns.xfr.TransferError:
            l.error('?loadZone: {} AXFR failed with NS={}'.format(self.name, ns))
            runner.send_msg('?loadZone {} failed. Giving up.'.format(self.name))
        except BaseException:
            l.error('%loadZone: {} Error on AXFR with NS={}, \n because {} - {}'.format(self.name,
                                                                                        ns,
                                                                                        sys.exc_info()[0].__name__,
                                                                                        str(sys.exc_info()[1])))
            runner.send_msg('%loadZone: {} Error on AXFR with NS={}, because {} - {}'.format(self.name,
                                                                                        ns,
                                                                                        sys.exc_info()[0].__name__,
                                                                                        str(sys.exc_info()[1])))

        if not ok:
            l.error('?loadZone {} failed. Giving up.'.format(self.name))
            runner.send_msg('?loadZone {} failed. Giving up.'.format(self.name))
            self.unreachable = True
            return

        err = ''
        try:
            self.z.check_origin()
        except dns.zone.NoSOA:
            err = 'SOA RRset'
        except dns.zone.NoNS:
            err = 'NS RRset'
        except KeyError:
            err = 'origin node'
        if err:
            l.error('% {} has no {}'.format(self.name, err))
            runner.send_msg('% {} has no {}'.format(self.name, err))

        l.debug('[loadZone: zone={}, AXFR of {} nodes done'.format(self.name, len(self.z.keys())))
        for k in self.z.keys():
            zn = name = str(k)
            if name == '@': name = ''
            self.d[row][1] = name
            node = self.z[zn]
            for the_rdataset in node:
                self.d[row][2] = str(the_rdataset.ttl)
                for rdata in the_rdataset:
                    srdatatype = dns.rdatatype.to_text(the_rdataset.rdtype)
                    if srdatatype not in ('RRSIG', 'NSEC', 'NSEC3'):
                        self.d[row][3] = srdatatype
                        srdata = str(rdata)
                        self.d[row][4] = srdata
                        try:
                            if srdatatype == 'MX':
                                srdata = str(rdata.exchange)
                            elif srdatatype == 'PTR':
                                a = dns.reversename.to_address(k)
                                ##if zad.prefs.debug: print('{}  {}'.format(a, name))
                            elif srdatatype == 'SOA' and srdata.endswith('.arpa.'):
                                a = dns.reversename.to_address(self.name)
                                ##if zad.prefs.debug: print('{}  {}'.format(a, self.name))
                        except dns.exception.SyntaxError:
                            l.error('%Malformed rdata: {} {} {}'.format(name, srdatatype, srdata))
                            runner.send_msg('%Malformed rdata: {} {} {}'.format(name, srdatatype, srdata))

                        else:
                            await self.createZoneFromName(srdatatype, srdata)
                        row = row + 1
                        self.d.append(['', '', '', '', ''])             # host(unused), name, ttl, type, rdata
                    ##l.debug('.', end=' ', flush=True)
                ##l.debug('+', end=' ', flush=True)
        self.valid = True
        l.info('[{} RRs out of {} nodes from zone {} parsed and linked.]'.format(
                                                            row -2,
                                                            len(self.z.keys()),
                                                            self.name))
        if zad.prefs.debug: logZones()
        runner.send_msg('Loading of {} completed with {} RRs'.format(self.name, row -2))
        runner.send_zone_loaded(self.name)


class Ip4Zone(Zone):
    def __init__(self, zone_name):
        self.init_by_subclass = True
        self.type = zad.common.ZTIP4
        super(Ip4Zone, self).__init__(zone_name)


class Ip6Zone(Zone):
    def __init__(self, zone_name):
        self.init_by_subclass = True
        self.type = zad.common.ZTIP6
        super(Ip6Zone, self).__init__(zone_name)


async def loadZones(runner: RunThread):
    """
    Check for zones which not have been loaded and load them
    """
    first_run_done = False
    def find_next_zone():
        zone_list = []
        for d in (domainZones, ip4Zones, ip6Zones):
            for k in d.keys():
                z = d[k]
                ## if len(z.d) < 2:
                if not z.valid and not z.unreachable and not z.ignored:
                    zone_list.append((z))
        if zone_list:
            return zone_list[0]
        else:
            return None

    while True:
        z = find_next_zone()
        if z:
            ## if first_run_done:              # zone has been modified by zad
            ##     await asyncio.sleep(10)      # allow for notify/axfr delay
            await z.loadZone(runner)
        else:
            if not first_run_done:
                logZones()
                runner.send_msg('All zones loaded.')
                first_run_done = True
            time.sleep(2)

def logZones():
    l.info('[Known Zones:]')
    pd = ''
    i = 0
    dtn = ['Domain Zones', 'IP4 Zones', 'IP6 Zones']
    
    for d in (domainZones, ip4Zones, ip6Zones):
        if not pd == d:
            l.info('\n[{}]'.format(dtn[i]))
            pd = d
            i += 1
        for k in d.keys():
            z = d[k]
            l.info('{}{}'.format(k, '[loaded]' if z.valid else ''))
            for n in z.nets.keys():
                l.info('        {}'.format(n))

def setupResolver(resolver_addresses):

    if not resolver_addresses:
        return
    
    dns.resolver.get_default_resolver()
    dns.resolver.default_resolver.nameservers = resolver_addresses
    dns.resolver.default_resolver.cache = dns.resolver.LRUCache()


def setup():
    setupResolver(zad.common.IP_RESOLVER)



import logging
import sys
import time
import dns.query, dns.resolver, dns.zone, dns.versioned, dns.rdataset, dns.rdatatype
import dns.exception
import dns.asyncbackend, dns.asyncquery
from PyQt5 import QtCore

import zad.common

l = logging.getLogger(__name__)
dns.asyncbackend.set_default_backend('asyncio')

domainZones = {}
ip4Zones = {}
ip6Zones = {}

subdomains = {}
ip4Nets = {}
ip6Nets = {}


class ZoneModel(QtCore.QAbstractTableModel):

    def __init__(self, data=[[]], netZone=False, parent=None):
        super(ZoneModel, self).__init__(parent)
        self.vrrs = data
        self.netZone = netZone

    def headerData(self, section: int, orientation: QtCore.Qt.Orientation, role: int):
        if role == QtCore.Qt.DisplayRole:
            if orientation == QtCore.Qt.Horizontal:
                f1 = 'Host' if self.netZone else 'OwnerName'
                return [f1, 'Type', 'Rdata'][section]
            else:
                return None

    def columnCount(self, parent=None):
        return 3

    def rowCount(self, parent=None):
        return len(self.vrrs) -1

    def data(self, index: QtCore.QModelIndex, role: int):
        if role == QtCore.Qt.DisplayRole:
            row = index.row()
            column = index.column()
            if self.netZone:
                if column > 0:
                    column += 2
            else:
                column += 1
                if column > 1:
                    column += 1
            v = self.vrrs[row][column]
            if not v:
                v = ''
            return str(v)

class EditZoneModel(ZoneModel):
    def __init__(self, data=[[]], netZone=False, parent=None):
        super(EditZoneModel, self).__init__(parent)
        self.vrrs = data
        self.netZone = netZone

    def headerData(self, section: int, orientation: QtCore.Qt.Orientation, role: int):
        if role == QtCore.Qt.DisplayRole:
            if orientation == QtCore.Qt.Horizontal:
                return ['Host', 'OwnerName', 'TTL', 'Type', 'Rdata'][section]
            else:
                return None

    def columnCount(self, parent=None):
        return 5

    def data(self, index: QtCore.QModelIndex, role: int):
        if role == QtCore.Qt.DisplayRole:
            row = index.row()
            column = index.column()
            v = self.vrrs[row][column]
            if not v:
                v = ''
            return str(v)

    """
    # keep column width with header
    def columnWidth(self) -> list(tuple):
        l = []
        if self netZone:
            l.append(0,20)
            l.append(1,100)
            l.append(2,10)
            l.append(3,10)
            l.append(4,400)
        else:
            l.append(0,120)
            l.append(1,10)
            l.append(2,10)
            l.append(3,400)
        return l
    """